#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import Dict, List, Union

import numpy as np
import scipy.ndimage as sc
import skimage.measure as skim


def get_matrix(roi_only_int: np.ndarray,
                     mask: np.ndarray,
                     levels: Union[np.ndarray, List]) -> np.ndarray:
    r"""
    Computes Grey level distance zone matrix.
    This matrix refers to "Grey level distance zone based features" (ID = VMDZ)  
    in the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_. 

    Args:
        roi_only_int (ndarray): 3D volume, isotropically resampled,
            quantized (e.g. n_g = 32, levels = [1, ..., n_g]),
            with NaNs outside the region of interest.
        mask (ndarray): Morphological ROI ``mask``.
        levels (ndarray or List): Vector containing the quantized gray-levels
                                  in the tumor region (or reconstruction ``levels`` of quantization).

    Returns:
        ndarray: Grey level distance zone Matrix.

    Todo:
        ``levels`` should be removed at some point, no longer needed if we always 
        quantize our volume such that ``levels = 1,2,3,4,...,max(quantized Volume)``.
        So simply calculate ``levels = 1:max(roi_only(~isnan(roi_only(:))))``
        directly in this function.

    """

    roi_only_int = roi_only_int.copy()
    levels = levels.copy().astype("int")
    morph_voxel_grid = mask.copy().astype(np.uint8)

    # COMPUTATION OF DISTANCE MAP
    morph_voxel_grid = np.pad(morph_voxel_grid,
                            [1,1],
                            'constant',
                            constant_values=0)

    # Computing the smallest ROI edge possible.
    # Distances are determined in 3D
    binary_struct = sc.generate_binary_structure(rank=3, connectivity=1)
    perimeter = morph_voxel_grid - sc.binary_erosion(morph_voxel_grid, structure=binary_struct)
    perimeter = perimeter[1:-1,1:-1,1:-1] # Removing the padding.
    morph_voxel_grid = morph_voxel_grid[1:-1,1:-1,1:-1] # Removing the padding

    # +1 according to the definition of the IBSI    
    dist_map = sc.distance_transform_cdt(np.logical_not(perimeter), metric='cityblock') + 1

    # INITIALIZATION
    # Since levels is always defined as 1,2,3,4,...,max(quantized Volume)
    n_g = np.size(levels)
    level_temp = np.max(levels) + 1
    roi_only_int[np.isnan(roi_only_int)] = level_temp
    # Since the ROI morph always encompasses ROI int,
    # using the mask as defined from ROI morph does not matter since
    # we want to find the maximal possible distance.
    dist_init = np.max(dist_map[morph_voxel_grid == 1])
    gldzm = np.zeros((n_g,dist_init))

    # COMPUTATION OF gldzm
    temp = roi_only_int.copy().astype('int')
    for i in range(1,n_g+1):
        temp[roi_only_int!=levels[i-1]] = 0
        temp[roi_only_int==levels[i-1]] = 1
        conn_objects, n_zone = skim.label(temp,return_num = True)
        for j in range(1,n_zone+1):
            col = np.min(dist_map[conn_objects==j]).astype("int")
            gldzm[i-1,col-1] = gldzm[i-1,col-1] + 1

    # REMOVE UNECESSARY COLUMNS
    stop = np.nonzero(np.sum(gldzm,0))[0][-1]
    gldzm = np.delete(gldzm, range(stop+1, np.shape(gldzm)[1]), 1)

    return  gldzm

def extract_all(vol_int: np.ndarray,
                mask_morph: np.ndarray,
                gldzm: np.ndarray = None) -> Dict:
    """Computes gldzm features.
    This feature refers to "Grey level distance zone based features" (ID = VMDZ)  
    in the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol_int (np.ndarray): 3D volume, isotropically resampled, quantized (e.g. n_g = 32, levels = [1, ..., n_g]),
        with NaNs outside the region of interest.
        mask_morph (np.ndarray): Morphological ROI mask.
        gldzm (np.ndarray, optional): array of the gray level distance zone matrix. Defaults to None.

    Returns:
        Dict: dict of ``gldzm`` features
    """
    gldzm_features = {'Fdzm_sde': [],
                      'Fdzm_lde': [],
                      'Fdzm_lgze': [],
                      'Fdzm_hgze': [],
                      'Fdzm_sdlge': [],
                      'Fdzm_sdhge': [],
                      'Fdzm_ldlge': [],
                      'Fdzm_ldhge': [],
                      'Fdzm_glnu': [],
                      'Fdzm_glnu_norm': [],
                      'Fdzm_zdnu': [],
                      'Fdzm_zdnu_norm': [],
                      'Fdzm_z_perc': [],
                      'Fdzm_gl_var': [],
                      'Fdzm_zd_var': [],
                      'Fdzm_zd_entr': []}

    # Correct definition, without any assumption
    levels = np.arange(1, np.max(vol_int[~np.isnan(vol_int[:])])+1)

    # GET THE gldzm MATRIX
    if gldzm is None:
        gldzm = get_matrix(vol_int, mask_morph, levels)
    n_s = np.sum(gldzm)
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    s_z = np.shape(gldzm)  # Size of gldzm
    c_vect = range(1, s_z[1]+1)  # Row vectors
    r_vect = range(1, s_z[0]+1)  # Column vectors
    # Column and row indicators for each entry of the gldzm
    c_mat, r_mat = np.meshgrid(c_vect, r_vect)
    p_g = np.transpose(np.sum(gldzm, 1))  # Gray-Level Vector
    p_d = np.sum(gldzm, 0)  # Distance Zone Vector

    # COMPUTING TEXTURES

    # Small distance emphasis
    gldzm_features['Fdzm_sde'] = (np.matmul(p_d, np.transpose(np.power(1.0/np.array(c_vect), 2))))

    # Large distance emphasis
    gldzm_features['Fdzm_lde'] = (np.matmul(p_d, np.transpose(np.power(np.array(c_vect), 2))))

    # Low grey level zone emphasis
    gldzm_features['Fdzm_lgze'] = np.matmul(p_g, np.transpose(np.power(1.0/np.array(r_vect), 2)))

    # High grey level zone emphasis
    gldzm_features['Fdzm_hgze'] = np.matmul(p_g, np.transpose(np.power(np.array(r_vect), 2)))

    # Small distance low grey level emphasis
    gldzm_features['Fdzm_sdlge'] = np.sum(np.sum(gldzm*(np.power(1.0/r_mat, 2))*(np.power(1.0/c_mat, 2))))

    # Small distance high grey level emphasis
    gldzm_features['Fdzm_sdhge'] = np.sum(np.sum(gldzm*(np.power(r_mat, 2))*(np.power(1.0/c_mat, 2))))

    # Large distance low grey level emphasis
    gldzm_features['Fdzm_ldlge'] = np.sum(np.sum(gldzm*(np.power(1.0/r_mat, 2))*(np.power(c_mat, 2))))

    # Large distance high grey level emphasis
    gldzm_features['Fdzm_ldhge'] = np.sum(np.sum(gldzm*(np.power(r_mat, 2))*(np.power(c_mat, 2))))

    # Gray level non-uniformity
    gldzm_features['Fdzm_glnu'] = np.sum(np.power(p_g, 2)) * n_s

    # Gray level non-uniformity normalised
    gldzm_features['Fdzm_glnu_norm'] = np.sum(np.power(p_g, 2))

    # Zone distance non-uniformity
    gldzm_features['Fdzm_zdnu'] = np.sum(np.power(p_d, 2)) * n_s

    # Zone distance non-uniformity normalised
    gldzm_features['Fdzm_zdnu_norm'] = np.sum(np.power(p_d, 2))

    # Zone percentage
    # Must change the original definition here.
    gldzm_features['Fdzm_z_perc'] = n_s / np.sum(~np.isnan(vol_int[:]))

    # Grey level variance
    temp = r_mat * gldzm
    u = np.sum(temp)
    temp = (np.power(r_mat-u, 2)) * gldzm
    gldzm_features['Fdzm_gl_var'] = np.sum(temp)

    # Zone distance variance
    temp = c_mat * gldzm
    u = np.sum(temp)
    temp = (np.power(c_mat-u, 2)) * gldzm
    temp = (np.power(c_mat - u, 2)) * gldzm
    gldzm_features['Fdzm_zd_var'] = np.sum(temp)

    # Zone distance entropy
    val_pos = gldzm[np.nonzero(gldzm)]
    temp = val_pos * np.log2(val_pos)
    gldzm_features['Fdzm_zd_entr'] = -np.sum(temp)

    return gldzm_features

def get_single_matrix(vol_int: np.ndarray, mask_morph: np.ndarray) -> np.ndarray:
    """Computes gray level distance zone matrix in order to compute the single features.

    Args:
        vol_int (ndarray): 3D volume, isotropically resampled,
                           quantized (e.g. n_g = 32, levels = [1, ..., n_g]),
                           with NaNs outside the region of interest.
        mask_morph (ndarray): Morphological ROI mask.

    Returns:
        ndarray: gldzm features.
    """
    # Correct definition, without any assumption
    levels = np.arange(1, np.max(vol_int[~np.isnan(vol_int[:])])+1)

    # GET THE gldzm MATRIX
    gldzm = get_matrix(vol_int, mask_morph, levels)

    return gldzm

def sde(gldzm: np.ndarray) -> float:
    """Computes small distance emphasis feature.
    This feature refers to "Fdzm_sde" (ID = 0GBI) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.
    
    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the small distance emphasis feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    s_z = np.shape(gldzm)  # Size of gldzm
    c_vect = range(1, s_z[1]+1)  # Row vectors
    p_d = np.sum(gldzm, 0)  # Distance Zone Vector

    # Small distance emphasis
    return (np.matmul(p_d, np.transpose(np.power(1.0 / np.array(c_vect), 2))))

def lde(gldzm: np.ndarray) -> float:
    """Computes large distance emphasis feature.
    This feature refers to "Fdzm_lde" (ID = MB4I) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the large distance emphasis feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    s_z = np.shape(gldzm)  # Size of gldzm
    c_vect = range(1, s_z[1]+1)  # Row vectors
    p_d = np.sum(gldzm, 0)  # Distance Zone Vector

    #Large distance emphasis
    return (np.matmul(p_d, np.transpose(np.power(np.array(c_vect), 2))))

def lgze(gldzm: np.ndarray) -> float:
    """Computes distance matrix low grey level zone emphasis feature.
    This feature refers to "Fdzm_lgze" (ID = S1RA) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the low grey level zone emphasis feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    s_z = np.shape(gldzm)  # Size of gldzm
    r_vect = range(1, s_z[0]+1)  # Column vectors
    p_g = np.transpose(np.sum(gldzm, 1))  # Gray-Level Vector

    #Low grey level zone emphasisphasis
    return np.matmul(p_g, np.transpose(np.power(1.0/np.array(r_vect), 2)))

def hgze(gldzm: np.ndarray) -> float:
    """Computes distance matrix high grey level zone emphasis feature.
    This feature refers to "Fdzm_hgze" (ID = K26C) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the high grey level zone emphasis feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    s_z = np.shape(gldzm)  # Size of gldzm
    r_vect = range(1, s_z[0]+1)  # Column vectors
    p_g = np.transpose(np.sum(gldzm, 1))  # Gray-Level Vector

    #Low grey level zone emphasisphasis
    return np.matmul(p_g, np.transpose(np.power(np.array(r_vect), 2)))

def sdlge(gldzm: np.ndarray) -> float:
    """Computes small distance low grey level emphasis feature.
    This feature refers to "Fdzm_sdlge" (ID = RUVG) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the low grey level emphasis feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    s_z = np.shape(gldzm)  # Size of gldzm
    c_vect = range(1, s_z[1]+1)  # Row vectors
    r_vect = range(1, s_z[0]+1)  # Column vectors
    c_mat, r_mat = np.meshgrid(c_vect, r_vect)  # Column and row indicators for each entry of the gldzm

    #Low grey level zone emphasisphasis
    return np.sum(np.sum(gldzm*(np.power(1.0/r_mat, 2))*(np.power(1.0/c_mat, 2))))

def sdhge(gldzm: np.ndarray) -> float:
    """Computes small distance high grey level emphasis feature.
    This feature refers to "Fdzm_sdhge" (ID = DKNJ) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the distance high grey level emphasis feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    s_z = np.shape(gldzm)  # Size of gldzm
    c_vect = range(1, s_z[1]+1)  # Row vectors
    r_vect = range(1, s_z[0]+1)  # Column vectors
    c_mat, r_mat = np.meshgrid(c_vect, r_vect)  # Column and row indicators for each entry of the gldzm

    #High grey level zone emphasisphasis
    return np.sum(np.sum(gldzm*(np.power(r_mat, 2))*(np.power(1.0/c_mat, 2))))

def ldlge(gldzm: np.ndarray) -> float:
    """Computes large distance low grey level emphasis feature.
    This feature refers to "Fdzm_ldlge" (ID = A7WM) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the low grey level emphasis feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    s_z = np.shape(gldzm)  # Size of gldzm
    c_vect = range(1, s_z[1]+1)  # Row vectors
    r_vect = range(1, s_z[0]+1)  # Column vectors
    c_mat, r_mat = np.meshgrid(c_vect, r_vect)  # Column and row indicators for each entry of the gldzm

    #Large distance low grey levels emphasis
    return np.sum(np.sum(gldzm*(np.power(1.0/r_mat, 2))*(np.power(c_mat, 2))))

def ldhge(gldzm: np.ndarray) -> float:
    """Computes large distance high grey level emphasis feature.
    This feature refers to "Fdzm_ldhge" (ID = KLTH) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the high grey level emphasis feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    s_z = np.shape(gldzm)  # Size of gldzm
    c_vect = range(1, s_z[1]+1)  # Row vectors
    r_vect = range(1, s_z[0]+1)  # Column vectors
    c_mat, r_mat = np.meshgrid(c_vect, r_vect)  # Column and row indicators for each entry of the gldzm

    #Large distance high grey levels emphasis
    return np.sum(np.sum(gldzm*(np.power(
        r_mat, 2))*(np.power(c_mat, 2))))

def glnu(gldzm: np.ndarray) -> float:
    """Computes distance zone matrix gray level non-uniformity
    This feature refers to "Fdzm_glnu" (ID = VFT7) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the gray level non-uniformity feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    p_g = np.transpose(np.sum(gldzm, 1))  # Gray-Level Vector
    n_s = np.sum(gldzm)

    #Gray level non-uniformity
    return np.sum(np.power(p_g, 2)) * n_s

def glnu_norm(gldzm: np.ndarray) -> float:
    """Computes distance zone matrix gray level non-uniformity normalised
    This feature refers to "Fdzm_glnu_norm" (ID = 7HP3) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the gray level non-uniformity normalised feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    p_g = np.transpose(np.sum(gldzm, 1))  # Gray-Level Vector

    #Gray level non-uniformity normalised
    return np.sum(np.power(p_g, 2))

def zdnu(gldzm: np.ndarray) -> float:
    """Computes zone distance non-uniformity
    This feature refers to "Fdzm_zdnu" (ID = V294) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the zone distance non-uniformity feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    p_d = np.sum(gldzm, 0)  # Distance Zone Vector
    n_s = np.sum(gldzm)

    #Zone distance non-uniformity
    return np.sum(np.power(p_d, 2)) * n_s

def zdnu_norm(gldzm: np.ndarray) -> float:
    """Computes zone distance non-uniformity normalised
    This feature refers to "Fdzm_zdnu_norm" (ID = IATH) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the zone distance non-uniformity normalised feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    p_d = np.sum(gldzm, 0)  # Distance Zone Vector

    #Zone distance non-uniformity normalised
    return np.sum(np.power(p_d, 2))

def z_perc(gldzm, vol_int):
    """Computes zone percentage
    This feature refers to "Fdzm_z_perc" (ID = VIWW) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the zone percentage feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    n_s = np.sum(gldzm)

    #Zone percentage
    return n_s/np.sum(~np.isnan(vol_int[:]))

def gl_var(gldzm: np.ndarray) -> float:
    """Computes grey level variance
    This feature refers to "Fdzm_gl_var" (ID = QK93) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the grey level variance feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    s_z = np.shape(gldzm)  # Size of gldzm
    c_vect = range(1, s_z[1]+1)  # Row vectors
    r_vect = range(1, s_z[0]+1)  # Column vectors
    _, r_mat = np.meshgrid(c_vect, r_vect)  # Column and row indicators for each entry of the gldzm
    temp = r_mat * gldzm
    u = np.sum(temp)
    temp = (np.power(r_mat-u, 2)) * gldzm

    #Grey level variance
    return np.sum(temp)

def zd_var(gldzm: np.ndarray) -> float:
    """Computes zone distance variance
    This feature refers to "Fdzm_zd_var" (ID = 7WT1) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the zone distance variance feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    s_z = np.shape(gldzm)  # Size of gldzm
    c_vect = range(1, s_z[1]+1)  # Row vectors
    r_vect = range(1, s_z[0]+1)  # Column vectors
    c_mat, _ = np.meshgrid(c_vect, r_vect)  # Column and row indicators for each entry of the gldzm
    temp = c_mat * gldzm
    u = np.sum(temp)
    temp = (np.power(c_mat-u, 2)) * gldzm

    #Zone distance variance
    return np.sum(temp)

def zd_entr(gldzm: np.ndarray) -> float:
    """Computes zone distance entropy
    This feature refers to "Fdzm_zd_entr" (ID = GBDU) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        gldzm (ndarray): array of the gray level distance zone matrix

    Returns:
        float: the zone distance entropy feature
    """
    gldzm = gldzm / np.sum(gldzm)  # Normalization of gldzm
    val_pos = gldzm[np.nonzero(gldzm)]
    temp = val_pos * np.log2(val_pos)

    #Zone distance entropy
    return -np.sum(temp)



from abc import ABC
from typing import Union

import numpy as np

from ..MEDscan import MEDscan
from ..utils.image_volume_obj import image_volume_obj
from .utils import convolve


class Mean():
    """The mean filter class"""

    def __init__(
                self,
                ndims: int,
                size: int,
                padding="symmetric"):
        """The constructor of the mean filter

        Args:
            ndims (int): Number of dimension of the kernel filter
            size (int): An integer that represent the length along one dimension of the kernel.
            padding: The padding type that will be used to produce the convolution

        Returns:
            None
        """

        assert isinstance(ndims, int) and ndims > 0, "ndims should be a positive integer"
        assert ((size+1)/2).is_integer() and size > 0, "size should be a positive odd number."

        self.padding = padding
        self.dim = ndims
        self.size = int(size)
        self.create_kernel()

    def create_kernel(self):
        """This method construct the mean kernel using the parameters specified to the constructor.

        Returns:
            ndarray: The mean kernel as a numpy multidimensional array
        """

        # Initialize the kernel as tensor of zeros
        weight = 1 / np.prod(self.size ** self.dim)
        kernel = np.ones([self.size for _ in range(self.dim)]) * weight

        self.kernel = np.expand_dims(kernel, axis=(0, 1))

    def convolve(self,
                images: np.ndarray,
                orthogonal_rot: bool = False)-> np.ndarray:
        """Filter a given image using the LoG kernel defined during the construction of this instance.
    
        Args:
            images (ndarray): A n-dimensional numpy array that represent the images to filter
            orthogonal_rot (bool, optional): If true, the 3D images will be rotated over coronal, axial and sagittal axis

        Returns:
            ndarray: The filtered image
        """
        # Swap the second axis with the last, to convert image B, W, H, D --> B, D, H, W
        image = np.swapaxes(images, 1, 3)
        result = np.squeeze(convolve(self.dim, self.kernel, image, orthogonal_rot, self.padding), axis=1)
        return np.swapaxes(result, 1, 3)

def apply_mean(
            input_images: Union[np.ndarray, image_volume_obj],
            medscan: MEDscan = None,
            ndims: int = 3,
            size: int = 15,
            padding: str = "symmetric",
            orthogonal_rot: bool = False
            ) -> np.ndarray:
    """Apply the mean filter to the input image

    Args:
        input_images (ndarray): The images to filter.
        medscan (MEDscan, optional): The MEDscan object that will provide the filter parameters.
        ndims (int, optional): The number of dimensions of the input image.
        size (int, optional): The size of the kernel.
        padding (str, optional): The padding type that will be used to produce the convolution. 
            Check options here: `numpy.pad <https://numpy.org/doc/stable/reference/generated/numpy.pad.html>`__.
        orthogonal_rot (bool, optional): If true, the 3D images will be rotated over coronal, axial and sagittal axis.

    Returns:
        ndarray: The filtered image.
    """
    # Check if the input is a numpy array or a Image volume object
    spatial_ref = None
    if type(input_images) == image_volume_obj:
        spatial_ref = input_images.spatialRef
        input_images = input_images.data
    
    # Convert to shape : (B, W, H, D)
    input_images = np.expand_dims(input_images.astype(np.float64), axis=0) 
    
    if medscan:
        # Initialize filter class instance
        _filter = Mean(
                ndims=medscan.params.filter.mean.ndims,
                size=medscan.params.filter.mean.size,
                padding=medscan.params.filter.mean.padding
                )
        # Run convolution
        result = _filter.convolve(input_images, orthogonal_rot=medscan.params.filter.mean.orthogonal_rot)
    else:
        # Initialize filter class instance
        _filter = Mean(
                    ndims=ndims,
                    size=size,
                    padding=padding,
                )
        # Run convolution
        result = _filter.convolve(input_images, orthogonal_rot=orthogonal_rot)

    if spatial_ref:
        return image_volume_obj(np.squeeze(result), spatial_ref)
    else:
        return np.squeeze(result)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-



from typing import Dict, Tuple, Union

import numpy as np


def get_matrix(roi_only:np.ndarray,
                     levels:np.ndarray,
                     dist_correction: bool=False) -> Tuple[np.ndarray, np.ndarray]:
    """This function computes the Neighborhood Gray-Tone Difference Matrix
    (NGTDM) of the region of interest (ROI) of an input volume. The input
    volume is assumed to be isotropically resampled. The ngtdm is computed
    using 26-voxel connectivity. To account for discretization length
    differences, all averages around a center voxel are performed such that
    the neighbours at a distance of :math:`\sqrt{3}` voxels are given a weight of
    :math:`\sqrt{3}`, and the neighbours at a distance of :math:`\sqrt{2}` voxels are given a
    weight of :math:`\sqrt{2}`.
    This matrix refers to "Neighbourhood grey tone difference based features" (ID = IPET)  
    in the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Note:
        This function is compatible with 2D analysis (language not adapted in the text)

    Args:
        roi_only (ndarray): Smallest box containing the ROI, with the imaging data ready
            for texture analysis computations. Voxels outside the ROI are set to NaNs.
        levels (ndarray): Vector containing the quantized gray-levels in the tumor region
            (or reconstruction ``levels`` of quantization).
        dist_correction (bool, optional): Set this variable to true in order to use
            discretization length difference corrections as used by the `Institute of Physics and
            Engineering in Medicine <https://doi.org/10.1088/0031-9155/60/14/5471>`_.
            Set this variable to false to replicate IBSI results.

    Returns:
        Tuple[np.ndarray, np.ndarray]: 
            - ngtdm: Neighborhood Gray-Tone Difference Matrix of ``roi_only'``.
            - count_valid: Array of number of valid voxels used in the ngtdm computation.

    REFERENCES:
        [1] Amadasun, M., & King, R. (1989). Textural Features Corresponding to
        Textural Properties. IEEE Transactions on Systems Man and Cybernetics,
        19(5), 1264–1274.
    
    """

    # PARSING "dist_correction" ARGUMENT
    if type(dist_correction) is not bool:
        # The user did not input either "true" or "false",
        # so the default behavior is used.
        dist_correction = True  # By default

    # PRELIMINARY
    if np.size(np.shape(roi_only)) == 2:  # generalization to 2D inputs
        two_d = 1
    else:
        two_d = 0

    roi_only = np.pad(roi_only, [1, 1], 'constant', constant_values=np.NaN)

    # # QUANTIZATION EFFECTS CORRECTION
    # # In case (for example) we initially wanted to have 64 levels, but due to
    # # quantization, only 60 resulted.
    unique_vol = levels.astype('int')
    NL = np.size(levels)
    temp = roi_only.copy().astype('int')
    for i in range(1, NL+1):
        roi_only[temp == unique_vol[i-1]] = i

    # INTIALIZATION
    ngtdm = np.zeros(NL)
    count_valid = np.zeros(NL)

    # COMPUTATION OF ngtdm
    if two_d:
        indices = np.where(~np.isnan(np.reshape(
            roi_only, np.size(roi_only), order='F')))[0]
        pos_valid = np.unravel_index(indices, np.shape(roi_only), order='F')
        n_valid_temp = np.size(pos_valid[0])
        w4 = 1
        if dist_correction:
            # Weights given to different neighbors to correct
            # for discretization length differences
            w8 = 1/np.sqrt(2)
        else:
            w8 = 1

        weights = np.array([w8, w4, w8, w4, w4, w4, w8, w4, w8])

        for n in range(1, n_valid_temp+1):

            neighbours = roi_only[(pos_valid[0][n-1]-1):(pos_valid[0][n-1]+2),
                                 (pos_valid[1][n-1]-1):(pos_valid[1][n-1]+2)].copy()
            neighbours = np.reshape(neighbours, 9, order='F')
            neighbours = neighbours*weights
            value = neighbours[4].astype('int')
            neighbours[4] = np.NaN
            neighbours = neighbours/np.sum(weights[~np.isnan(neighbours)])
            neighbours = np.delete(neighbours, 4)  # Remove the center voxel
            # Thus only excluding voxels with NaNs only as neighbors.
            if np.size(neighbours[~np.isnan(neighbours)]) > 0:
                ngtdm[value-1] = ngtdm[value-1] + np.abs(
                    value-np.sum(neighbours[~np.isnan(neighbours)]))
                count_valid[value-1] = count_valid[value-1] + 1
    else:

        indices = np.where(~np.isnan(np.reshape(
            roi_only, np.size(roi_only), order='F')))[0]
        pos_valid = np.unravel_index(indices, np.shape(roi_only), order='F')
        n_valid_temp = np.size(pos_valid[0])
        w6 = 1
        if dist_correction:
            # Weights given to different neighbors to correct
            # for discretization length differences
            w26 = 1 / np.sqrt(3)
            w18 = 1 / np.sqrt(2)
        else:
            w26 = 1
            w18 = 1

        weights = np.array([w26, w18, w26, w18, w6, w18, w26, w18, w26, w18,
                            w6, w18, w6, w6, w6, w18, w6, w18, w26, w18,
                            w26, w18, w6, w18, w26, w18, w26])

        for n in range(1, n_valid_temp+1):
            neighbours = roi_only[(pos_valid[0][n-1]-1) : (pos_valid[0][n-1]+2),
                                 (pos_valid[1][n-1]-1) : (pos_valid[1][n-1]+2),
                                 (pos_valid[2][n-1]-1) : (pos_valid[2][n-1]+2)].copy()
            neighbours = np.reshape(neighbours, 27, order='F')
            neighbours = neighbours * weights
            value = neighbours[13].astype('int')
            neighbours[13] = np.NaN
            neighbours = neighbours / np.sum(weights[~np.isnan(neighbours)])
            neighbours = np.delete(neighbours, 13)  # Remove the center voxel
            # Thus only excluding voxels with NaNs only as neighbors.
            if np.size(neighbours[~np.isnan(neighbours)]) > 0:
                ngtdm[value-1] = ngtdm[value-1] + np.abs(value - np.sum(neighbours[~np.isnan(neighbours)]))
                count_valid[value-1] = count_valid[value-1] + 1

    return ngtdm, count_valid

def extract_all(vol: np.ndarray,
                dist_correction :Union[bool, str]=None) -> Dict:
    """Compute Neighbourhood grey tone difference based features.
    These features refer to "Neighbourhood grey tone difference based features" (ID = IPET) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:

        vol (ndarray): 3D volume, isotropically resampled, quantized
                       (e.g. n_g = 32, levels = [1, ..., n_g]), with NaNs outside the region
                       of interest.
        dist_correction (Union[bool, str], optional): Set this variable to true in order to use
                                                      discretization length difference corrections as used
                                                      by the `Institute of Physics and Engineering in
                                                      Medicine <https://doi.org/10.1088/0031-9155/60/14/5471>`__.
                                                      Set this variable to false to replicate IBSI results.
                                                      Or use string and specify the norm for distance weighting.
                                                      Weighting is only performed if this argument is
                                                      "manhattan", "euclidean" or "chebyshev".
    
    Returns:
        Dict: Dict of Neighbourhood grey tone difference based features.
    """
    ngtdm_features = {'Fngt_coarseness': [],
             'Fngt_contrast': [],
             'Fngt_busyness': [],
             'Fngt_complexity': [],
             'Fngt_strength': []}

    # GET THE NGTDM MATRIX
    # Correct definition, without any assumption
    levels = np.arange(1, np.max(vol[~np.isnan(vol[:])].astype("int"))+1)
    ngtdm, count_valid = get_matrix(vol, levels, dist_correction)

    n_tot = np.sum(count_valid)
    # Now representing the probability of gray-level occurences
    count_valid = count_valid/n_tot
    nl = np.size(ngtdm)
    n_g = np.sum(count_valid != 0)
    p_valid = np.where(np.reshape(count_valid, np.size(
        count_valid), order='F') > 0)[0]+1
    n_valid = np.size(p_valid)

    # COMPUTING TEXTURES
    # Coarseness
    coarseness = 1 / np.matmul(np.transpose(count_valid), ngtdm)
    coarseness = min(coarseness, 10**6)
    ngtdm_features['Fngt_coarseness'] = coarseness

    # Contrast
    if n_g == 1:
        ngtdm_features['Fngt_contrast'] = 0
    else:
        val = 0
        for i in range(1, nl+1):
            for j in range(1, nl+1):
                val = val + count_valid[i-1] * count_valid[j-1] * ((i-j)**2)
        ngtdm_features['Fngt_contrast'] = val * np.sum(ngtdm) / (n_g*(n_g-1)*n_tot)

    # Busyness
    if n_g == 1:
        ngtdm_features['Fngt_busyness'] = 0
    else:
        denom = 0
        for i in range(1, n_valid+1):
            for j in range(1, n_valid+1):
                denom = denom + np.abs(p_valid[i-1]*count_valid[p_valid[i-1]-1] -
                                       p_valid[j-1]*count_valid[p_valid[j-1]-1])
        ngtdm_features['Fngt_busyness'] = np.matmul(np.transpose(count_valid), ngtdm) / denom

    # Complexity
    val = 0
    for i in range(1, n_valid+1):
        for j in range(1, n_valid+1):
            val = val + (np.abs(
                p_valid[i-1]-p_valid[j-1]) / (n_tot*(
                count_valid[p_valid[i-1]-1] +
                count_valid[p_valid[j-1]-1])))*(
                count_valid[p_valid[i-1]-1]*ngtdm[p_valid[i-1]-1] +
                count_valid[p_valid[j-1]-1]*ngtdm[p_valid[j-1]-1])

    ngtdm_features['Fngt_complexity'] = val

    # Strength
    if np.sum(ngtdm) == 0:
        ngtdm_features['Fngt_strength'] = 0
    else:
        val = 0
        for i in range(1, n_valid+1):
            for j in range(1, n_valid+1):
                val = val + (count_valid[p_valid[i-1]-1] + count_valid[p_valid[j-1]-1])*(
                    p_valid[i-1]-p_valid[j-1])**2

        ngtdm_features['Fngt_strength'] = val/np.sum(ngtdm)

    return ngtdm_features

def get_single_matrix(vol: np.ndarray,
               dist_correction = None)-> Tuple[np.ndarray,
                                               np.ndarray]:
    """Compute neighbourhood grey tone difference matrix in order to compute the single features.

    Args:

        vol (ndarray): 3D volume, isotropically resampled, quantized
            (e.g. n_g = 32, levels = [1, ..., n_g]), with NaNs outside the region of interest.
        dist_correction (Union[bool, str], optional): Set this variable to true in order to use
                                                      discretization length difference corrections as used
                                                      by the `Institute of Physics and Engineering in
                                                      Medicine <https://doi.org/10.1088/0031-9155/60/14/5471>`__.
                                                      Set this variable to false to replicate IBSI results.
                                                      Or use string and specify the norm for distance weighting.
                                                      Weighting is only performed if this argument is
                                                      "manhattan", "euclidean" or "chebyshev".
    
    Returns:
        np.ndarray: array of neighbourhood grey tone difference matrix
    """
    # GET THE NGTDM MATRIX
    # Correct definition, without any assumption
    levels = np.arange(1, np.max(vol[~np.isnan(vol[:])].astype("int"))+1)

    ngtdm, count_valid = get_matrix(vol, levels, dist_correction)
    
    return ngtdm, count_valid

def coarseness(ngtdm: np.ndarray, count_valid: np.ndarray)-> float:
    """
    Computes coarseness feature.
    This feature refers to "Coarseness" (ID = QCDE) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngtdm (ndarray): array of neighbourhood grey tone difference matrix
    
    Returns:
        float: coarseness value

    """
    n_tot = np.sum(count_valid)
    count_valid = count_valid/n_tot
    coarseness = 1 / np.matmul(np.transpose(count_valid), ngtdm)
    coarseness = min(coarseness, 10**6)
    return coarseness

def contrast(ngtdm: np.ndarray, count_valid: np.ndarray)-> float:
    """
    Computes contrast feature.
    This feature refers to "Contrast" (ID = 65HE) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngtdm (ndarray): array of neighbourhood grey tone difference matrix
    
    Returns:
        float: contrast value

    """
    n_tot = np.sum(count_valid)
    count_valid = count_valid/n_tot
    nl = np.size(ngtdm)
    n_g = np.sum(count_valid != 0)

    if n_g == 1:
        return 0
    else:
        val = 0
        for i in range(1, nl+1):
            for j in range(1, nl+1):
                val = val + count_valid[i-1] * count_valid[j-1] * ((i-j)**2)
        contrast = val * np.sum(ngtdm) / (n_g*(n_g-1)*n_tot)
        return contrast

def busyness(ngtdm: np.ndarray, count_valid: np.ndarray)-> float:
    """
    Computes busyness feature.
    This feature refers to "Busyness" (ID = NQ30) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngtdm (ndarray): array of neighbourhood grey tone difference matrix
    
    Returns:
        float: busyness value

    """
    n_tot = np.sum(count_valid)
    count_valid = count_valid/n_tot
    n_g = np.sum(count_valid != 0)
    p_valid = np.where(np.reshape(count_valid, np.size(
        count_valid), order='F') > 0)[0]+1
    n_valid = np.size(p_valid)

    if n_g == 1:
        busyness = 0
        return busyness
    else:
        denom = 0
        for i in range(1, n_valid+1):
            for j in range(1, n_valid+1):
                denom = denom + np.abs(p_valid[i-1]*count_valid[p_valid[i-1]-1] -
                                       p_valid[j-1]*count_valid[p_valid[j-1]-1])
        busyness = np.matmul(np.transpose(count_valid), ngtdm) / denom 
        return busyness

def complexity(ngtdm: np.ndarray, count_valid: np.ndarray)-> float:
    """
    Computes complexity feature.
    This feature refers to "Complexity" (ID = HDEZ) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngtdm (ndarray): array of neighbourhood grey tone difference matrix
    
    Returns:
        float: complexity value

    """
    n_tot = np.sum(count_valid)
    # Now representing the probability of gray-level occurences
    count_valid = count_valid/n_tot
    p_valid = np.where(np.reshape(count_valid, np.size(
        count_valid), order='F') > 0)[0]+1
    n_valid = np.size(p_valid)

    val = 0
    for i in range(1, n_valid+1):
        for j in range(1, n_valid+1):
            val = val + (np.abs(
                p_valid[i-1]-p_valid[j-1]) / (n_tot*(
                count_valid[p_valid[i-1]-1] +
                count_valid[p_valid[j-1]-1])))*(
                count_valid[p_valid[i-1]-1]*ngtdm[p_valid[i-1]-1] +
                count_valid[p_valid[j-1]-1]*ngtdm[p_valid[j-1]-1])
    complexity = val
    return complexity

def strength(ngtdm: np.ndarray, count_valid: np.ndarray)-> float:
    """
    Computes strength feature.
    This feature refers to "Strength" (ID = 1X9X) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        ngtdm (ndarray): array of neighbourhood grey tone difference matrix
    
    Returns:
        float: strength value

    """

    n_tot = np.sum(count_valid)
    # Now representing the probability of gray-level occurences
    count_valid = count_valid/n_tot
    p_valid = np.where(np.reshape(count_valid, np.size(
        count_valid), order='F') > 0)[0]+1
    n_valid = np.size(p_valid)

    if np.sum(ngtdm) == 0:
        strength = 0
        return strength
    else:
        val = 0
        for i in range(1, n_valid+1):
            for j in range(1, n_valid+1):
                val = val + (count_valid[p_valid[i-1]-1] + count_valid[p_valid[j-1]-1])*(
                    p_valid[i-1]-p_valid[j-1])**2

        strength = val/np.sum(ngtdm)
        return strength



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from re import finditer
from typing import List


def strfind(pattern: str,
            string: str) -> List[int]:
    """Finds indices of ``pattern`` in ``string``. Based on regex.

    Note:
        Be careful with + and - symbols. Use :math:`\+` and :math:`\-` instead.

    Args:
        pattern (str): Substring to be searched in the ``string``.
        string (str): String used to find matches.

    Returns:
        List[int]: List of indexes of every occurence of ``pattern`` in the passed ``string``.

    Raises:
        ValueError: If the ``pattern`` does not use backslash with special regex symbols
    """

    if pattern in ('+', '-'):
        raise ValueError(
            "Please use a backslash with special regex symbols in findall.")

    ind = [m.start() for m in finditer(pattern, string)]

    return ind



import random
from typing import List

import numpy as np
import pandas as pd


def check_min_n_per_cat(
        variable_table: pd.DataFrame, 
        var_names: List[str], 
        min_n_per_cat: float, 
        type: str) -> pd.DataFrame:
    """
    This Function is different from matlab, it takes the whole variable_table
    and the name of the var_of_type to fit the way pandas works

    Args:
        variable_table (pd.DataFrame): Table of variables.
        var_names (list): List of variable names.
        min_n_per_cat (float): Minimum number of observations per category.
        type (str): Type of variable.
    
    Returns:
        pd.DataFrame: Table of variables with categories under ``min_n_per_cat``.
    """

    for name in var_names:
        table = variable_table[var_names]
        cats = pd.Categorical(table[name]).categories
        for cat in cats:
            flag_cat = (table == cat)
            if sum(flag_cat[name]) < min_n_per_cat:
                if type == 'hcategorical':
                    table.mask(flag_cat, np.nan)
                if type == 'icategorical':
                    table.mask(flag_cat, '')
        variable_table[var_names] = table

    return variable_table

def check_max_percent_cat(variable_table, var_names, max_percent_cat) -> pd.Series:
    """
    This Function is different from matlab, it takes the whole variable_table
    and the name of the var_of_type to fit the way pandas works

    Args:
        variable_table (pd.DataFrame): Table of variables.
        var_names (list): List of variable names.
        max_percent_cat (float): Maximum number of observations per category.
    
    Returns:
        pd.DataFrame: Table of variables with categories over ``max_percent_cat``.
    """

    n_observation = variable_table.shape[0]
    flag_var_out = pd.Series(np.zeros(var_names.size, dtype=bool))
    n = 0
    for name in var_names:
        cats = pd.Categorical(variable_table[name]).categories
        for cat in cats:
            if (variable_table[name] == cat).sum()/n_observation > max_percent_cat:
                flag_var_out[n] = True
                break
        n += 1
    return flag_var_out

def one_hot_encode_table(variable_table: pd.DataFrame) -> pd.DataFrame:
    """
    Converts a table of categorical variables into a table of one-hot encoded variables.

    Args:
        variable_table (pd.DataFrame): Table of variables.
    
    Returns:
        variable_table (pd.DataFrame): Table of variables with one-hot encoded variables.
    """

    #INITIALIZATION
    var_icat = variable_table.Properties['userData']['variables']['icategorical']
    n_var_icat = var_icat.size
    if n_var_icat == 0:
        return variable_table

    # ONE-HOT ENCODING
    for var_name in var_icat:
        categories = variable_table[var_name].unique()
        categories = np.asarray(list(filter(lambda v: v == v, categories)))  # get rid of nan
        categories.sort()
        n_categories = categories.size
        name_encoded = []
        position_to_add = variable_table.columns.get_loc(var_name)+1
        if n_categories == 2:
            n_categories = 1
        for c in range(n_categories):
            cat = categories[c]
            new_name = f"{var_name}__{cat}"
            data_to_add = (variable_table[var_name] == cat).astype(int)
            variable_table.insert(loc=position_to_add, column=new_name, value=data_to_add)
            name_encoded.append(new_name)
        variable_table.Properties['userData']['variables']["one_hot"] = dict()
        variable_table.Properties['userData']['variables']["one_hot"][var_name] = name_encoded
        variable_table = variable_table.drop(var_name, axis=1)

    # UPDATING THE VARIABLE TYPES
    variable_table.Properties['userData']['variables']["icategorical"] = np.array([])
    variable_table.Properties['userData']['variables']["hcategorical"] = np.append([], name_encoded)
    return variable_table



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy as np


def inpolygon(x_q: np.ndarray,
              y_q: np.ndarray,
              x_v: np.ndarray,
              y_v: np.ndarray) -> np.ndarray:
    """Implements similar functionality MATLAB inpolygon.
    Finds points located inside or on edge of polygonal region.

    Note: 
        Unlike matlab inpolygon, this function does not determine the
        status of single points :math:`(x_q, y_q)`. Instead, it determines the 
        status for an entire grid by ray-casting.

    Args:
        x_q (ndarray): x-coordinates of query points, in intrinsic reference system.
        y_q (ndarray): y-coordinates of query points, in intrinsic reference system.
        x_q (ndarray): x-coordinates of polygon vertices, in intrinsic reference system.
        y_q (ndarray): y-coordinates of polygon vertices, in intrinsic reference system.

    Returns:
        ndarray: boolean array indicating if the query points are on the edge of the polygon area.

    """
    def ray_line_intersection(ray_orig, ray_dir, vert_1, vert_2):
        """
        
        """
        epsilon = 0.000001

        # Define edge
        edge_line = vert_1 - vert_2

        # Define ray vertices
        r_vert_1 = ray_orig
        r_vert_2 = ray_orig + ray_dir
        edge_ray = - ray_dir

        # Calculate determinant - if close to 0, lines are parallel and will
        # not intersect
        det = np.cross(edge_ray, edge_line)
        if (det > -epsilon) and (det < epsilon):
            return np.nan

        # Calculate inverse of the determinant
        inv_det = 1.0 / det

        # Calculate determinant
        a11 = np.cross(r_vert_1, r_vert_2)
        a21 = np.cross(vert_1, vert_2)

        # Solve for x
        a12 = edge_ray[0]
        a22 = edge_line[0]
        x = np.linalg.det(np.array([[a11, a12], [a21, a22]])) * inv_det

        # Solve for y
        b12 = edge_ray[1]
        b22 = edge_line[1]
        y = np.linalg.det(np.array([[a11, b12], [a21, b22]])) * inv_det

        t = np.array([x, y])

        # Check whether the solution falls within the line segment
        u1 = np.around(np.dot(edge_line, edge_line), 5)
        u2 = np.around(np.dot(edge_line, vert_1-t), 5)
        if (u2 / u1) < 0.0 or (u2 / u1) > 1.0:
            return np.nan

        # Return scalar length from ray origin
        t_scal = np.linalg.norm(ray_orig - t)

        return t_scal

    # These are hacks to actually make this function work
    spacing = np.array([1.0, 1.0])
    origin = np.array([0.0, 0.0])
    shape = np.array([np.max(x_q) + 1, np.max(y_q) + 1])
    # shape = np.array([np.max(x_q), np.max(y_q)]) Original from Alex
    vertices = np.vstack((x_v, y_v)).transpose()
    lines = np.vstack(
        ([np.arange(0, len(x_v))], [np.arange(-1, len(x_v) - 1)])).transpose()

    # Set up line vertices
    vertex_a = vertices[lines[:, 0], :]
    vertex_b = vertices[lines[:, 1], :]

    # Remove lines with length 0 and center on the origin
    line_mask = np.sum(np.abs(vertex_a - vertex_b), axis=1) > 0.0
    vertex_a = vertex_a[line_mask] - origin
    vertex_b = vertex_b[line_mask] - origin

    # Find extent of contours in x
    x_min_ind = int(
        np.max([np.floor(np.min(vertices[:, 0]) / spacing[0]), 0.0]))
    x_max_ind = int(
        np.min([np.ceil(np.max(vertices[:, 0]) / spacing[0]), shape[0] * 1.0]))

    # Set up voxel grid and y-span
    vox_grid = np.zeros(shape, dtype=int)
    vox_span = origin[1] + np.arange(0, shape[1]) * spacing[1]

    # Set ray origin and direction (starts at negative y, and travels towards
    # positive y
    ray_origin = np.array([0.0, -1.0])
    ray_dir = np.array([0.0, 1.0])

    for x_ind in np.arange(x_min_ind, x_max_ind):
        # Update ray origin
        ray_origin[0] = origin[0] + x_ind * spacing[0]

        # Scan both forward and backward to resolve points located on
        # the polygon
        vox_col_frwd = np.zeros(np.shape(vox_span), dtype=int)
        vox_col_bkwd = np.zeros(np.shape(vox_span), dtype=int)

        # Find lines that are intersected by the ray
        ray_hit = np.sum(
            np.sign(np.vstack((vertex_a[:, 0], vertex_b[:, 0])) - ray_origin[0]), axis=0)

        # If the ray crosses a vertex, the sum of the sign is 0 when the ray
        # does not hit an vertex point, and -1 or 1 when it does.
        # In the latter case, we only keep of the vertices for each hit.
        simplex_mask = np.logical_or(ray_hit == 0, ray_hit == 1)

        # Go to next iterator if mask is empty
        if np.sum(simplex_mask) == 0:
            continue

        # Determine the selected vertices
        selected_verts = np.squeeze(np.where(simplex_mask))

        # Find intercept of rays with lines
        t_scal = np.array([ray_line_intersection(ray_orig=ray_origin, ray_dir=ray_dir,
                                                 vert_1=vertex_a[ii, :], vert_2=vertex_b[ii, :]) for ii in selected_verts])

        # Remove invalid results
        t_scal = t_scal[np.isfinite(t_scal)]
        if t_scal.size == 0:
            continue

        # Update vox_col based on t_scal. This basically adds a 1 for all
        # voxels that lie behind the line intersections
        # of the ray.
        for t_curr in t_scal:
            vox_col_frwd[vox_span > t_curr + ray_origin[1]] += 1
        for t_curr in t_scal:
            vox_col_bkwd[vox_span < t_curr + ray_origin[1]] += 1

        # Voxels in the roi cross an uneven number of meshes from the origin
        vox_grid[x_ind,
                 :] += np.logical_and(vox_col_frwd % 2, vox_col_bkwd % 2)

    return vox_grid.astype(dtype=bool)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from copy import deepcopy
from typing import Tuple

import numpy as np
from skimage.exposure import equalize_hist


def equalization(vol_re: np.ndarray) -> np.ndarray:
    """Performs histogram equalisation of the ROI imaging intensities.

    Note:
        This is a pure "what is contained within the roi" equalization. this is
        not influenced by the :func:`user_set_min_val()` used for FBS discretisation.

    Args:
        vol_re (ndarray): 3D array of the image volume that will be studied with
                          NaN value for the excluded voxels (voxels outside the ROI mask).

    Returns:
        ndarray: Same input image volume but with redistributed intensities.
    """

    # AZ: This was made part of the function call
    # n_g = 64
    # This is the default we will use. It means that when using 'FBS',
    # n_q should be chosen wisely such
    # that the total number of grey levels does not exceed 64, for all
    # patients (recommended).
    # This choice was amde by considering that the best equalization
    # performance for "histeq.m" is obtained with low n_g.
    # WARNING: The effective number of grey levels coming out of "histeq.m"
    # may be lower than n_g.

    # CONSERVE THE INDICES OF THE ROI
    x_gl = np.ravel(vol_re)
    ind_roi = np.where(~np.isnan(vol_re))
    x_gl = x_gl[~np.isnan(x_gl)]

    # ADJUST RANGE BETWEEN 0 and 1
    min_val = np.min(x_gl)
    max_val = np.max(x_gl)
    x_gl_01 = (x_gl - min_val)/(max_val - min_val)

    # EQUALIZATION
    # x_gl_equal = equalize_hist(x_gl_01, nbins=n_g)
    # AT THE MOMENT, WE CHOOSE TO USE THE DEFAULT NUMBER OF BINS OF
    # equalize_hist.py (256)
    x_gl_equal = equalize_hist(x_gl_01)
    # RE-ADJUST TO CORRECT RANGE
    x_gl_equal = (x_gl_equal - np.min(x_gl_equal)) / \
        (np.max(x_gl_equal) - np.min(x_gl_equal))
    x_gl_equal = x_gl_equal * (max_val - min_val)
    x_gl_equal = x_gl_equal + min_val

    # RECONSTRUCT THE VOLUME WITH EQUALIZED VALUES
    vol_equal_re = deepcopy(vol_re)

    vol_equal_re[ind_roi] = x_gl_equal

    return vol_equal_re

def discretize(vol_re: np.ndarray,
                   discr_type: str,
                   n_q: float=None,
                   user_set_min_val: float=None,
                   ivh=False) -> Tuple[np.ndarray, float]:
    """Quantisizes the image intensities inside the ROI.

    Note:
        For 'FBS' type, it is assumed that re-segmentation with
        proper range was already performed

    Args:
        vol_re (ndarray): 3D array of the image volume that will be studied with
                          NaN value for the excluded voxels (voxels outside the ROI mask).
        discr_type (str): Discretisaion approach/type must be: "FBS", "FBN", "FBSequal"
                          or "FBNequal".
        n_q (float): Number of bins for FBS algorithm and bin width for FBN algorithm.
        user_set_min_val (float): Minimum of range re-segmentation for FBS discretisation,
                                  for FBN discretisation, this value has no importance as an argument
                                  and will not be used.
        ivh (bool): Must be set to True for IVH (Intensity-Volume histogram) features.

    Returns:
        2-element tuple containing

        - ndarray: Same input image volume but with discretised intensities.
        - float: bin width.
    """

    # AZ: NOTE: the "type" variable that appeared in the MATLAB source code
    # matches the name of a standard python function. I have therefore renamed
    # this variable "discr_type"

    # PARSING ARGUMENTS
    vol_quant_re = deepcopy(vol_re)

    if n_q is None:
        return None

    if not isinstance(n_q, float):
        n_q = float(n_q)

    if discr_type not in ["FBS", "FBN", "FBSequal", "FBNequal"]:
        raise ValueError(
            "discr_type must either be \"FBS\", \"FBN\", \"FBSequal\" or \"FBNequal\".")

    # DISCRETISATION
    if discr_type in ["FBS", "FBSequal"]:
        if user_set_min_val:
            min_val = deepcopy(user_set_min_val)
        else:
            min_val = np.nanmin(vol_quant_re)
    else:
        min_val = np.nanmin(vol_quant_re)

    max_val = np.nanmax(vol_quant_re)

    if discr_type == "FBS":
        w_b = n_q
        w_d = w_b
        vol_quant_re = np.floor((vol_quant_re - min_val) / w_b) + 1.0
    elif discr_type == "FBN":
        w_b = (max_val - min_val) / n_q
        w_d = 1.0
        vol_quant_re = np.floor(
            n_q * ((vol_quant_re - min_val)/(max_val - min_val))) + 1.0
        vol_quant_re[vol_quant_re == np.nanmax(vol_quant_re)] = n_q
    elif discr_type == "FBSequal":
        w_b = n_q
        w_d = w_b
        vol_quant_re = equalization(vol_quant_re)
        vol_quant_re = np.floor((vol_quant_re - min_val) / w_b) + 1.0
    elif discr_type == "FBNequal":
        w_b = (max_val - min_val) / n_q
        w_d = 1.0
        vol_quant_re = vol_quant_re.astype(np.float32)
        vol_quant_re = equalization(vol_quant_re)
        vol_quant_re = np.floor(
            n_q * ((vol_quant_re - min_val)/(max_val - min_val))) + 1.0
        vol_quant_re[vol_quant_re == np.nanmax(vol_quant_re)] = n_q
    if ivh and discr_type in ["FBS", "FBSequal"]:
        vol_quant_re = min_val + (vol_quant_re - 0.5) * w_b

    return vol_quant_re, w_d



import math
from itertools import product
from typing import Union

import numpy as np

from ..MEDscan import MEDscan
from ..utils.image_volume_obj import image_volume_obj
from .utils import convolve


class LaplacianOfGaussian():
    """The Laplacian of gaussian filter class."""

    def __init__(
            self,
            ndims: int,
            size: int,
            sigma: float=0.1,
            padding: str="symmetric"):
        """The constructor of the laplacian of gaussian (LoG) filter

        Args:
            ndims (int): Number of dimension of the kernel filter
            size (int): An integer that represent the length along one dimension of the kernel.
            sigma (float): The gaussian standard deviation parameter of the laplacian of gaussian filter
            padding (str): The padding type that will be used to produce the convolution

        Returns:
            None
        """

        assert isinstance(ndims, int) and ndims > 0, "ndims should be a positive integer"
        assert ((size+1)/2).is_integer() and size > 0, "size should be a positive odd number."
        assert sigma > 0, "alpha should be a positive float."

        self.dim = ndims
        self.padding = padding
        self.size = int(size)
        self.sigma = sigma
        self.create_kernel()

    def create_kernel(self) -> np.ndarray:
        """This method construct the LoG kernel using the parameters specified to the constructor
    
        Returns:
            ndarray: The laplacian of gaussian kernel as a numpy multidimensional array
        """

        def compute_weight(position):
            distance_2 = np.sum(position**2)
            # $\frac{-1}{\sigma^2} * \frac{1}{\sqrt{2 \pi} \sigma}^D = \frac{-1}{\sqrt{D/2}{2 \pi} * \sigma^{D+2}}$
            first_part = -1/((2*math.pi)**(self.dim/2) * self.sigma**(self.dim+2))

            # $(D - \frac{||k||^2}{\sigma^2}) * e^{\frac{-||k||^2}{2 \sigma^2}}$
            second_part = (self.dim - distance_2/self.sigma**2)*math.e**(-distance_2/(2 * self.sigma**2))

            return first_part * second_part

        # Initialize the kernel as tensor of zeros
        kernel = np.zeros([self.size for _ in range(self.dim)])

        for k in product(range(self.size), repeat=self.dim):
            kernel[k] = compute_weight(np.array(k)-int((self.size-1)/2))

        kernel -= np.sum(kernel)/np.prod(kernel.shape)
        self.kernel = np.expand_dims(kernel, axis=(0, 1))

    def convolve(self,
                 images: np.ndarray,
                 orthogonal_rot=False) -> np.ndarray:
        """Filter a given image using the LoG kernel defined during the construction of this instance.

        Args:
            images (ndarray): A n-dimensional numpy array that represent the images to filter
            orthogonal_rot (bool): If true, the 3D images will be rotated over coronal, axial and sagittal axis

        Returns:
            ndarray: The filtered image
        """
        # Swap the second axis with the last, to convert image B, W, H, D --> B, D, H, W
        image = np.swapaxes(images, 1, 3)
        result = np.squeeze(convolve(self.dim, self.kernel, image, orthogonal_rot, self.padding), axis=1)
        return np.swapaxes(result, 1, 3)

def apply_log(
        input_images: Union[np.ndarray, image_volume_obj],
        medscan: MEDscan = None,
        ndims: int = 3,
        voxel_length: float = 0.0,
        sigma: float = 0.1,
        padding: str = "symmetric",
        orthogonal_rot: bool = False
    ) -> np.ndarray:
    """Apply the mean filter to the input image

    Args:
        input_images (ndarray): The images to filter.
        medscan (MEDscan, optional): The MEDscan object that will provide the filter parameters.
        ndims (int, optional): The number of dimensions of the input image.
        voxel_length (float, optional): The voxel size of the input image.
        sigma (float, optional): standard deviation of the Gaussian, controls the scale of the convolutional operator.
        padding (str, optional): The padding type that will be used to produce the convolution. 
            Check options here: `numpy.pad <https://numpy.org/doc/stable/reference/generated/numpy.pad.html>`__.
        orthogonal_rot (bool, optional): If true, the 3D images will be rotated over coronal, axial and sagittal axis.

    Returns:
        ndarray: The filtered image.
    """
    # Check if the input is a numpy array or a Image volume object
    spatial_ref = None
    if type(input_images) == image_volume_obj:
        spatial_ref = input_images.spatialRef
        input_images = input_images.data
    
    # Convert to shape : (B, W, H, D)
    input_images = np.expand_dims(input_images.astype(np.float64), axis=0) 
    
    if medscan:
        # Initialize filter class params & instance
        sigma = medscan.params.filter.log.sigma / voxel_length
        length = 2 * int(4 * sigma + 0.5) + 1
        _filter = LaplacianOfGaussian(
                    ndims=medscan.params.filter.log.ndims,
                    size=length,
                    sigma=sigma,
                    padding=medscan.params.filter.log.padding
                )
        # Run convolution
        result = _filter.convolve(input_images, orthogonal_rot=medscan.params.filter.log.orthogonal_rot)
    else:
        # Initialize filter class params & instance
        sigma = sigma / voxel_length
        length = 2 * int(4 * sigma + 0.5) + 1
        _filter = LaplacianOfGaussian(
                    ndims=ndims,
                    size=length,
                    sigma=sigma,
                    padding=padding
                )
        # Run convolution
        result = _filter.convolve(input_images, orthogonal_rot=orthogonal_rot)

    if spatial_ref:
        return image_volume_obj(np.squeeze(result), spatial_ref)
    else:
        return np.squeeze(result)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from importlib.resources import path
from pathlib import Path
from typing import List, Union


def get_file_paths(path_to_parent_folder: Union[str, Path], wildcard: str=None) -> List[Path]:
    """Finds all files in the given path that matches the pattern/wildcard. 

    Note:
        The search is done recursively in all subdirectories.

    Args:
        path_to_parent_folder (Union[str, Path]): Full path to where the files are located.
        wildcard (str, optional): String specifying which type of files 
        to locate in the parent folder.
            - Ex : '*.dcm*', to look for dicom files.

    Returns:
        List: List of full paths to files with the specific wildcard located \
            in the given path to parent folder.
    """
    if wildcard is None:
        wildcard = '*'

    # Getting the list of all files full path in file_paths
    path_to_parent_folder = Path(path_to_parent_folder)
    file_paths_list = list(path_to_parent_folder.rglob(wildcard))
    # for the name only put file.name
    file_paths = [file for file in file_paths_list if file.is_file()]

    return file_paths



from . import *
from .cleaning_utils import *
from .DataCleaner import DataCleaner
from .DesignExperiment import DesignExperiment
from .FSR import FSR
from .ml_utils import *
from .Normalization import Normalization
from .RadiomicsLearner import RadiomicsLearner
from .Results import Results
from .Stats import Stats



import numpy as np

from ..MEDscan import MEDscan
from ..utils.image_volume_obj import image_volume_obj
from .gabor import *
from .laws import *
from .log import *
from .mean import *
try:
    from .TexturalFilter import TexturalFilter
except ImportError:
    import_failed = True
from .wavelet import *


def apply_filter(
            medscan: MEDscan,
            vol_obj: Union[image_volume_obj, np.ndarray],
            user_set_min_val: float = None,
            feature: str = None
    ) -> Union[image_volume_obj, np.ndarray]:
    """Applies mean filter on the given data

    Args:
        medscan (MEDscan): Instance of the MEDscan class that holds the filtering params
        vol_obj (image_volume_obj): Imaging data to be filtered
        user_set_min_val (float, optional): The minimum value to use for the discretization. Defaults to None.
        feature (str, optional): The feature to extract from the family. In batch extraction, all the features 
            of the family will be extracted. Defaults to None.

    Returns:
        image_volume_obj: Filtered imaging data.
    """
    filter_type = medscan.params.filter.filter_type

    if filter_type.lower() == "mean":
        input = np.expand_dims(vol_obj.data.astype(np.float64), axis=0)   # Convert to shape : (B, W, H, D)
        # Initialize filter class instance
        _filter = Mean(
                    ndims=medscan.params.filter.mean.ndims,
                    size=medscan.params.filter.mean.size,
                    padding=medscan.params.filter.mean.padding
                )
        # Run convolution
        result = _filter.convolve(input, orthogonal_rot=medscan.params.filter.mean.orthogonal_rot)

    elif filter_type.lower() == "log":
        # Initialize filter class params & instance
        input = np.expand_dims(vol_obj.data.astype(np.float64), axis=0)   # Convert to shape : (B, W, H, D)
        voxel_length = medscan.params.process.scale_non_text[0]
        sigma = medscan.params.filter.log.sigma / voxel_length
        length = 2 * int(4 * sigma + 0.5) + 1
        _filter = LaplacianOfGaussian(
                    ndims=medscan.params.filter.log.ndims,
                    size=length,
                    sigma=sigma,
                    padding=medscan.params.filter.log.padding
                )
        # Run convolution
        result = _filter.convolve(input, orthogonal_rot=medscan.params.filter.log.orthogonal_rot)

    elif filter_type.lower() == "laws":
        # Initialize filter class instance
        input = np.expand_dims(vol_obj.data.astype(np.float64), axis=0)   # Convert to shape : (B, W, H, D)
        _filter = Laws(
                    config=medscan.params.filter.laws.config, 
                    energy_distance=medscan.params.filter.laws.energy_distance,
                    rot_invariance=medscan.params.filter.laws.rot_invariance,
                    padding=medscan.params.filter.laws.padding
                )
        # Run convolution
        result = _filter.convolve(
                    input, 
                    orthogonal_rot=medscan.params.filter.laws.orthogonal_rot,
                    energy_image=medscan.params.filter.laws.energy_image
                )

    elif filter_type.lower() == "gabor":
        # Initialize filter class params & instance
        input = np.expand_dims(vol_obj.data.astype(np.float64), axis=0)   # Convert to shape : (B, W, H, D)
        voxel_length = medscan.params.process.scale_non_text[0]
        sigma = medscan.params.filter.gabor.sigma / voxel_length
        lamb = medscan.params.filter.gabor._lambda / voxel_length
        size = 2 * int(7 * sigma + 0.5) + 1
        _filter = Gabor(size=size,
                        sigma=sigma,
                        lamb=lamb,
                        gamma=medscan.params.filter.gabor.gamma,
                        theta=-medscan.params.filter.gabor.theta,
                        rot_invariance=medscan.params.filter.gabor.rot_invariance,
                        padding=medscan.params.filter.gabor.padding
                        )
        # Run convolution
        result = _filter.convolve(input, orthogonal_rot=medscan.params.filter.gabor.orthogonal_rot)

    elif filter_type.lower().startswith("wavelet"):
        # Initialize filter class instance
        input = np.expand_dims(vol_obj.data.astype(np.float64), axis=0)   # Convert to shape : (B, W, H, D)
        _filter = Wavelet(
                        ndims=medscan.params.filter.wavelet.ndims, 
                        wavelet_name=medscan.params.filter.wavelet.basis_function,
                        rot_invariance=medscan.params.filter.wavelet.rot_invariance,
                        padding=medscan.params.filter.wavelet.padding
                    )
        # Run convolution
        result = _filter.convolve(
                        input, 
                        _filter=medscan.params.filter.wavelet.subband, 
                        level=medscan.params.filter.wavelet.level
                    )
    elif filter_type.lower() == "textural":
        if not import_failed:
            # Initialize filter class instance
            _filter = TexturalFilter(
                family=medscan.params.filter.textural.family,
            )
            # Apply filter
            vol_obj = _filter(
                vol_obj,
                size=medscan.params.filter.textural.size,
                discretization=medscan.params.filter.textural.discretization,
                local=medscan.params.filter.textural.local,
                user_set_min_val=user_set_min_val,
                feature=feature
            )
    else:
        raise ValueError(
                r'Filter name should either be: "mean", "log", "laws", "gabor" or "wavelet".'
                )

    if not filter_type.lower() == "textural":
        vol_obj.data = np.squeeze(result,axis=0)

    return vol_obj



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import Dict, List, Tuple


def initialize_features_names(image_space_struct: Dict) -> Tuple[List, List]:
    """Finds all the features names from `image_space_struct`

    Args:
        image_space_struct(Dict): Dictionary of the extracted features (Texture & Non-texture)
    
    Returns:
        Tuple[List, List]: Two lists of the texture and non-texture features names found in the `image_space_struct`.
    """
    # First entry is the names of feature types. Second entry is the name of
    # the features for a given feature type. Third entry is the name of the
    # extraction parameters for all features of a given feature type.
    non_text_cell = [0] * 3
    # First entry is the names of feature types. Second entry is the name of
    # the features for a given feature type. Third entry is the name of the
    # extraction parameters for all features of a given feature type.
    text_cell = [0] * 3

    # NON-TEXTURE FEATURES
    field_non_text = [key for key in image_space_struct.keys() if key != 'texture']
    n_non_text_type = len(field_non_text)
    non_text_cell[0] = field_non_text
    non_text_cell[1] = [0] * n_non_text_type
    non_text_cell[2] = [0] * n_non_text_type

    for t in range(0, n_non_text_type):
        dic_image_space_struct_non_text = image_space_struct[non_text_cell[0][t]]
        field_params_non_text = [
            key for key in dic_image_space_struct_non_text.keys()]
        dic_image_space_struct_params_non_text = image_space_struct[non_text_cell[0]
                                                               [t]][field_params_non_text[0]]
        field_feat_non_text = [
            key for key in dic_image_space_struct_params_non_text.keys()]
        non_text_cell[1][t] = field_feat_non_text
        non_text_cell[2][t] = field_params_non_text

    # TEXTURE FEATURES
    dic_image_space_struct_texture = image_space_struct['texture']
    field_text = [key for key in dic_image_space_struct_texture.keys()]
    n_text_type = len(field_text)
    text_cell[0] = field_text
    text_cell[1] = [0] * n_text_type
    text_cell[2] = [0] * n_text_type

    for t in range(0, n_text_type):
        dic_image_space_struct_text = image_space_struct['texture'][text_cell[0][t]]
        field_params_text = [key for key in dic_image_space_struct_text.keys()]
        dic_image_space_struct_params_text = image_space_struct['texture'][text_cell[0]
                                                                       [t]][field_params_text[0]]
        field_feat_text = [
            key for key in dic_image_space_struct_params_text.keys()]
        text_cell[1][t] = field_feat_text
        text_cell[2][t] = field_params_text

    return non_text_cell, text_cell



from typing import List

import numpy as np
from scipy.signal import fftconvolve


def pad_imgs(
            images: np.ndarray,
            padding_length: List,
            axis: List,
            mode: str
            )-> np.ndarray:
    """Apply padding on a 3d images using a 2D padding pattern.

    Args:
        images (ndarray): a numpy array that represent the image.
        padding_length (List): The padding length that will apply on each side of each axe.
        axis (List): A list of axes on which the padding will be done.
        mode (str): The padding mode. Check options here: `numpy.pad 
            <https://numpy.org/doc/stable/reference/generated/numpy.pad.html>`__.

    Returns:
        ndarray: A numpy array that represent the padded image.
    """
    pad_tuple = ()
    j = 1

    for i in range(np.ndim(images)):
        if i in axis:
            pad_tuple += ((padding_length[-j], padding_length[-j]),)
            j += 1
        else:
            pad_tuple += ((0, 0),)

    return np.pad(images, pad_tuple, mode=mode)

def convolve(
        dim: int,
        kernel: np.ndarray,
        images: np.ndarray,
        orthogonal_rot: bool=False,
        mode: str = "symmetric"
    ) -> np.ndarray:
    """Convolve a given n-dimensional array with the kernel to generate a filtered image.

    Args:
        dim (int): The dimension of the images.
        kernel (ndarray): The kernel to use for the convolution.
        images (ndarray): A n-dimensional numpy array that represent a batch of images to filter.
        orthogonal_rot (bool, optional): If true, the 3D images will be rotated over coronal, axial and sagittal axis.
        mode (str, optional): The padding mode. Check options here: `numpy.pad 
            <https://numpy.org/doc/stable/reference/generated/numpy.pad.html>`__.

    Returns:
        ndarray: The filtered image.
    """

    in_size = np.shape(images)

    # We only handle 2D or 3D images.
    assert len(in_size) == 3 or len(in_size) == 4, \
        "The tensor should have the followed shape (B, H, W) or (B, D, H, W)"

    if not orthogonal_rot:
        # If we have a 2D kernel but a 3D images, we squeeze the tensor
        if dim < len(in_size) - 1:
            images = images.reshape((in_size[0] * in_size[1], in_size[2], in_size[3]))

        # We compute the padding size along each dimension
        padding = [int((kernel.shape[-1] - 1) / 2) for _ in range(dim)]
        pad_axis_list = [i for i in range(1, dim+1)]

        # We pad the images and we add the channel axis.
        padded_imgs = pad_imgs(images, padding, pad_axis_list, mode)
        new_imgs = np.expand_dims(padded_imgs, axis=1)

        # Operate the convolution
        if dim < len(in_size) - 1:
            # If we have a 2D kernel but a 3D images, we convolve slice by slice
            result_list = [fftconvolve(np.expand_dims(new_imgs[i], axis=0), kernel, mode='valid') for i in range(len(images))]
            result = np.squeeze(np.stack(result_list), axis=2)

        else :
            result = fftconvolve(new_imgs, kernel, mode='valid')

        # Reshape the data to retrieve the following format: (B, C, D, H, W)
        if dim < len(in_size) - 1:
            result = result.reshape((
                in_size[0], in_size[1], result.shape[1], in_size[2], in_size[3])
            ).transpose(0, 2, 1, 3, 4)

    # If we want orthogonal rotation
    else:
        coronal_imgs = images
        axial_imgs, sagittal_imgs = np.rot90(images, 1, (1, 2)), np.rot90(images, 1, (1, 3))
        
        result_coronal = convolve(dim, kernel, coronal_imgs, False, mode)
        result_axial = convolve(dim, kernel, axial_imgs, False, mode)
        result_sagittal = convolve(dim, kernel, sagittal_imgs, False, mode)

        # split and unflip and stack the result on a new axis
        result_axial = np.rot90(result_axial, 1, (3, 2))
        result_sagittal = np.rot90(result_sagittal, 1, (4, 2))

        result = np.stack([result_coronal, result_axial, result_sagittal])

    return result



import logging
import os
from json import dump
from pathlib import Path
from typing import Dict, List, Union

import matplotlib.pyplot as plt
import nibabel as nib
import numpy as np
from numpyencoder import NumpyEncoder
from PIL import Image

from .utils.image_volume_obj import image_volume_obj
from .utils.imref import imref3d
from .utils.json_utils import load_json


class MEDscan(object):
    """Organizes all scan data (patientID, imaging data, scan type...). 

    Attributes:
        patientID (str): Patient ID.
        type (str): Scan type (MRscan, CTscan...).
        format (str): Scan file format. Either 'npy' or 'nifti'.
        dicomH (pydicom.dataset.FileDataset): DICOM header.
        data (MEDscan.data): Instance of MEDscan.data inner class.

    """

    def __init__(self, medscan=None) -> None:
        """Constructor of the MEDscan class

        Args:
            medscan(MEDscan): A MEDscan class instance.
        
        Returns:
            None
        """
        try:
            self.patientID = medscan.patientID
        except:
            self.patientID = ""
        try:
            self.type = medscan.type
        except:
            self.type = ""
        try:
            self.series_description = medscan.series_description
        except:
            self.series_description = ""
        try:
            self.format = medscan.format
        except:
            self.format = ""
        try:
            self.dicomH = medscan.dicomH
        except:
            self.dicomH = []
        try:
            self.data = medscan.data
        except:
            self.data = self.data()

        self.params = self.Params()
        self.radiomics = self.Radiomics()
        self.skip = False

    def __init_process_params(self, im_params: Dict) -> None:
        """Initializes the processing params from a given Dict.
        
        Args:
            im_params(Dict): Dictionary of different processing params.
        
        Returns:
            None.
        """
        if self.type == 'CTscan' and 'imParamCT' in im_params:
            im_params = im_params['imParamCT']
        elif self.type == 'MRscan' and 'imParamMR' in im_params:
            im_params = im_params['imParamMR']
        elif self.type == 'PTscan' and 'imParamPET' in im_params:
            im_params = im_params['imParamPET']
        else:
            raise ValueError(f"The given parameters dict is not valid, no params found for {self.type} modality")
        
        # re-segmentation range processing
        if(im_params['reSeg']['range'] and (im_params['reSeg']['range'][0] == "inf" or im_params['reSeg']['range'][0] == "-inf")):
            im_params['reSeg']['range'][0] = -np.inf
        if(im_params['reSeg']['range'] and im_params['reSeg']['range'][1] == "inf"):
            im_params['reSeg']['range'][1] = np.inf

        if 'box_string' in im_params:
            box_string = im_params['box_string']
        else:
            # By default, we add 10 voxels in all three dimensions are added to the smallest
            # bounding box. This setting is used to speed up interpolation
            # processes (mostly) prior to the computation of radiomics
            # features. Optional argument in the function computeRadiomics.
            box_string = 'box10'
        if 'compute_diag_features' in im_params:
            compute_diag_features = im_params['compute_diag_features']
        else:
            compute_diag_features = False
        if compute_diag_features:  # If compute_diag_features is true.
            box_string = 'full'  # This is required for proper comparison.

        self.params.process.box_string = box_string

        # get default scan parameters from im_param_scan
        self.params.process.scale_non_text = im_params['interp']['scale_non_text']
        self.params.process.vol_interp  = im_params['interp']['vol_interp']
        self.params.process.roi_interp = im_params['interp']['roi_interp']
        self.params.process.gl_round = im_params['interp']['gl_round']
        self.params.process.roi_pv  = im_params['interp']['roi_pv']
        self.params.process.im_range = im_params['reSeg']['range'] if 'range' in im_params['reSeg'] else None
        self.params.process.outliers = im_params['reSeg']['outliers']
        self.params.process.ih = im_params['discretisation']['IH']
        self.params.process.ivh = im_params['discretisation']['IVH']
        self.params.process.scale_text = im_params['interp']['scale_text']
        self.params.process.algo = im_params['discretisation']['texture']['type'] if 'type' in im_params['discretisation']['texture'] else []
        self.params.process.gray_levels = im_params['discretisation']['texture']['val'] if 'val' in im_params['discretisation']['texture'] else [[]]
        self.params.process.im_type = self.type

        # Voxels dimension
        self.params.process.n_scale = len(self.params.process.scale_text)
        # Setting up discretisation params
        self.params.process.n_algo = len(self.params.process.algo)
        self.params.process.n_gl = len(self.params.process.gray_levels[0])
        self.params.process.n_exp = self.params.process.n_scale * self.params.process.n_algo * self.params.process.n_gl

        # Setting up user_set_min_value
        if self.params.process.im_range is not None and type(self.params.process.im_range) is list and self.params.process.im_range:
            user_set_min_value = self.params.process.im_range[0]
            if user_set_min_value == -np.inf:
                # In case no re-seg im_range is defined for the FBS algorithm,
                # the minimum value of ROI will be used (not recommended).
                user_set_min_value = []
        else:
            # In case no re-seg im_range is defined for the FBS algorithm,
            # the minimum value of ROI will be used (not recommended).
            user_set_min_value = [] 
        self.params.process.user_set_min_value = user_set_min_value

        # box_string argument is optional. If not present, we use the full box.
        if self.params.process.box_string is None:
            self.params.process.box_string = 'full'
        
        # set filter type for the modality
        if 'filter_type' in im_params:
            self.params.filter.filter_type = im_params['filter_type']
        
        # Set intensity type
        if 'intensity_type' in im_params and im_params['intensity_type'] != "":
            self.params.process.intensity_type = im_params['intensity_type']
        elif self.params.filter.filter_type != "":
            self.params.process.intensity_type = 'filtered'
        elif self.type == 'MRscan':
            self.params.process.intensity_type = 'arbitrary'
        else:
            self.params.process.intensity_type = 'definite'

    def __init_extraction_params(self, im_params: Dict):
        """Initializes the extraction params from a given Dict.
        
        Args:
            im_params(Dict): Dictionary of different extraction params.
        
        Returns:
            None.
        """
        if self.type == 'CTscan' and 'imParamCT' in im_params:
            im_params = im_params['imParamCT']
        elif self.type == 'MRscan' and 'imParamMR' in im_params:
            im_params = im_params['imParamMR']
        elif self.type == 'PTscan' and 'imParamPET' in im_params:
            im_params = im_params['imParamPET']
        else:
            raise ValueError(f"The given parameters dict is not valid, no params found for {self.type} modality")
        
        # glcm features extraction params
        if 'glcm' in im_params:
            if 'dist_correction' in im_params['glcm']:
                self.params.radiomics.glcm.dist_correction = im_params['glcm']['dist_correction']
            else:
                self.params.radiomics.glcm.dist_correction = False
            if 'merge_method' in im_params['glcm']:
                self.params.radiomics.glcm.merge_method = im_params['glcm']['merge_method']
            else:
                self.params.radiomics.glcm.merge_method = "vol_merge"
        else:
            self.params.radiomics.glcm.dist_correction = False
            self.params.radiomics.glcm.merge_method = "vol_merge"

        # glrlm features extraction params
        if 'glrlm' in im_params:
            if 'dist_correction' in im_params['glrlm']:
                self.params.radiomics.glrlm.dist_correction = im_params['glrlm']['dist_correction']
            else:
                self.params.radiomics.glrlm.dist_correction = False
            if 'merge_method' in im_params['glrlm']:
                self.params.radiomics.glrlm.merge_method = im_params['glrlm']['merge_method']
            else:
                self.params.radiomics.glrlm.merge_method = "vol_merge"
        else:
            self.params.radiomics.glrlm.dist_correction = False
            self.params.radiomics.glrlm.merge_method = "vol_merge"


        # ngtdm features extraction params
        if 'ngtdm' in im_params:
            if 'dist_correction' in im_params['ngtdm']:
                self.params.radiomics.ngtdm.dist_correction = im_params['ngtdm']['dist_correction']
            else:
                self.params.radiomics.ngtdm.dist_correction = False
        else:
            self.params.radiomics.ngtdm.dist_correction = False
        
        # Features to extract
        features = [
            "Morph", "LocalIntensity", "Stats", "IntensityHistogram", "IntensityVolumeHistogram", 
            "GLCM", "GLRLM", "GLSZM", "GLDZM", "NGTDM", "NGLDM"
        ]
        if "extract" in im_params.keys():
            self.params.radiomics.extract = im_params['extract']
            for key in self.params.radiomics.extract:
                if key not in features:
                    raise ValueError(f"Invalid key in 'extract' parameter: {key} (Modality {self.type}).")
        
        # Ensure each feature is in the extract dictionary with a default value of True
        for feature in features:
            if feature not in self.params.radiomics.extract:
                self.params.radiomics.extract[feature] = True

    def __init_filter_params(self, filter_params: Dict) -> None:
        """Initializes the filtering params from a given Dict.

        Args:
            filter_params(Dict): Dictionary of the filtering parameters.
        
        Returns:
            None.
        """
        if 'imParamFilter' in filter_params:
            filter_params = filter_params['imParamFilter']
        
        # Initializae filter attribute
        self.params.filter = self.params.Filter()

        # mean filter params
        if 'mean' in filter_params:
            self.params.filter.mean.init_from_json(filter_params['mean'])

        # log filter params
        if 'log' in filter_params:
            self.params.filter.log.init_from_json(filter_params['log'])

        # laws filter params
        if 'laws' in filter_params:
            self.params.filter.laws.init_from_json(filter_params['laws'])

        # gabor filter params
        if 'gabor' in filter_params:
            self.params.filter.gabor.init_from_json(filter_params['gabor'])

        # wavelet filter params
        if 'wavelet' in filter_params:
            self.params.filter.wavelet.init_from_json(filter_params['wavelet'])

        # Textural filter params
        if 'textural' in filter_params:
            self.params.filter.textural.init_from_json(filter_params['textural'])

    def init_params(self, im_param_scan: Dict) -> None:
        """Initializes the Params class from a dictionary.

        Args:
            im_param_scan(Dict): Dictionary of different processing, extraction and filtering params.
        
        Returns:
            None.
        """
        try:
            # get default scan parameters from im_param_scan
            self.__init_filter_params(im_param_scan['imParamFilter'])
            self.__init_process_params(im_param_scan)
            self.__init_extraction_params(im_param_scan)

            # compute suv map for PT scans
            if self.type == 'PTscan':
                _compute_suv_map = im_param_scan['imParamPET']['compute_suv_map']
            else :
                _compute_suv_map = False
            
            if self.type == 'PTscan' and _compute_suv_map and self.format != 'nifti':
                try:
                    from .processing.compute_suv_map import compute_suv_map
                    self.data.volume.array = compute_suv_map(self.data.volume.array, self.dicomH[0])
                except Exception as e :
                    message = f"\n ERROR COMPUTING SUV MAP - SOME FEATURES WILL BE INVALID: \n {e}"
                    logging.error(message)
                    print(message)
                    self.skip = True
            
            # initialize radiomics structure
            self.radiomics.image = {}
            self.radiomics.params = im_param_scan
            self.params.radiomics.scale_name = ''
            self.params.radiomics.ih_name = ''
            self.params.radiomics.ivh_name = ''
            
        except Exception as e:
            message = f"\n ERROR IN INITIALIZATION OF RADIOMICS FEATURE COMPUTATION\n {e}"
            logging.error(message)
            print(message)
            self.skip = True

    def init_ntf_calculation(self, vol_obj: image_volume_obj) -> None:
        """
        Initializes all the computation parameters for non-texture features  as well as the results dict.

        Args:
            vol_obj(image_volume_obj): Imaging volume.
        
        Returns:
            None.
        """
        try:
            if sum(self.params.process.scale_non_text) == 0:  # In case the user chose to not interpolate
                self.params.process.scale_non_text = [
                                        vol_obj.spatialRef.PixelExtentInWorldX,
                                        vol_obj.spatialRef.PixelExtentInWorldY,
                                        vol_obj.spatialRef.PixelExtentInWorldZ]
            else:
                if len(self.params.process.scale_non_text) == 2:
                    # In case not interpolation is performed in
                    # the slice direction (e.g. 2D case)
                    self.params.process.scale_non_text = self.params.process.scale_non_text + \
                        [vol_obj.spatialRef.PixelExtentInWorldZ]

            # Scale name
            # Always isotropic resampling, so the first entry is ok.
            self.params.radiomics.scale_name = 'scale' + (str(self.params.process.scale_non_text[0])).replace('.', 'dot')

            # IH name
            if 'val' in self.params.process.ih:
                ih_val_name = 'bin' + (str(self.params.process.ih['val'])).replace('.', 'dot')
            else:
                ih_val_name = 'binNone'

            # The minimum value defines the computation.
            if self.params.process.ih['type'].find('FBS')>=0:
                if type(self.params.process.user_set_min_value) is list and self.params.process.user_set_min_value:
                    min_val_name = '_min' + \
                        ((str(self.params.process.user_set_min_value)).replace('.', 'dot')).replace('-', 'M')
                else:
                    # Otherwise, minimum value of ROI will be used (not recommended),
                    # so no need to report it.
                    min_val_name = ''
            else:
                min_val_name = ''
            self.params.radiomics.ih_name = self.params.radiomics.scale_name + \
                                            '_algo' + self.params.process.ih['type'] + \
                                            '_' + ih_val_name + min_val_name

            # IVH name
            if self.params.process.im_range:  # The im_range defines the computation.
                min_val_name = ((str(self.params.process.im_range[0])).replace('.', 'dot')).replace('-', 'M')
                max_val_name = ((str(self.params.process.im_range[1])).replace('.', 'dot')).replace('-', 'M')
                if max_val_name == 'inf':
                    # In this case, the maximum value of the ROI is used,
                    # so no need to report it.
                    range_name = '_min' + min_val_name
                elif min_val_name == '-inf' or min_val_name == 'inf':
                    # In this case, the minimum value of the ROI is used,
                    # so no need to report it.
                    range_name = '_max' + max_val_name
                else:
                    range_name = '_min' + min_val_name + '_max' + max_val_name
            else: 
                # min-max of ROI will be used, no need to report it.
                range_name = ''
            if not self.params.process.ivh:  # CT case for example
                ivh_algo_name = 'algoNone'
                ivh_val_name = 'bin1'
            else:
                ivh_algo_name = 'algo' + self.params.process.ivh['type'] if 'type' in self.params.process.ivh else 'algoNone'
                if 'val' in self.params.process.ivh and self.params.process.ivh['val']:
                    ivh_val_name = 'bin' + (str(self.params.process.ivh['val'])).replace('.', 'dot')
                else:
                    ivh_val_name = 'binNone'
            self.params.radiomics.ivh_name = self.params.radiomics.scale_name + '_' + ivh_algo_name + '_' + ivh_val_name + range_name

            # Now initialize the attribute that will hold the computation results
            self.radiomics.image.update({ 
                'morph_3D': {self.params.radiomics.scale_name: {}},
                'locInt_3D': {self.params.radiomics.scale_name: {}},
                'stats_3D': {self.params.radiomics.scale_name: {}},
                'intHist_3D': {self.params.radiomics.ih_name: {}},
                'intVolHist_3D': {self.params.radiomics.ivh_name: {}} 
            })

        except Exception as e:
            message = f"\n PROBLEM WITH PRE-PROCESSING OF FEATURES IN init_ntf_calculation(): \n {e}"
            logging.error(message)
            print(message)
            self.radiomics.image.update(
                    {('scale' + (str(self.params.process.scale_non_text[0])).replace('.', 'dot')): 'ERROR_PROCESSING'})

    def init_tf_calculation(self, algo:int, gl:int, scale:int) -> None:
        """
        Initializes all the computation parameters for the texture-features as well as the results dict.

        Args:
            algo(int): Discretisation algorithms index.
            gl(int): gray-level index.
            scale(int): scale-text index.
        
        Returns:
            None.
        """
        # check glcm merge method
        glcm_merge_method = self.params.radiomics.glcm.merge_method
        if glcm_merge_method:
            if glcm_merge_method == 'average':
                glcm_merge_method = '_avg'
            elif glcm_merge_method == 'vol_merge':
                glcm_merge_method = '_comb'
            else:
                error_msg = f"{glcm_merge_method} Method not supported in glcm computation, \
                    only 'average' or 'vol_merge' are supported. \
                    Radiomics will be saved without any specific merge method."
                logging.warning(error_msg)
                print(error_msg)

        # check glrlm merge method
        glrlm_merge_method = self.params.radiomics.glrlm.merge_method
        if glrlm_merge_method:
            if glrlm_merge_method == 'average':
                glrlm_merge_method = '_avg'
            elif glrlm_merge_method == 'vol_merge':
                glrlm_merge_method = '_comb'
            else:
                error_msg = f"{glcm_merge_method} Method not supported in glrlm computation, \
                    only 'average' or 'vol_merge' are supported. \
                    Radiomics will be saved without any specific merge method"
                logging.warning(error_msg)
                print(error_msg)
        # set texture features names and updates radiomics dict
        self.params.radiomics.name_text_types = [
                                'glcm_3D' + glcm_merge_method, 
                                'glrlm_3D' + glrlm_merge_method, 
                                'glszm_3D', 
                                'gldzm_3D', 
                                'ngtdm_3D', 
                                'ngldm_3D']
        n_text_types = len(self.params.radiomics.name_text_types)
        if not ('texture' in self.radiomics.image):
            self.radiomics.image.update({'texture': {}})
            for t in range(n_text_types):
                self.radiomics.image['texture'].update({self.params.radiomics.name_text_types[t]: {}})

        # scale name
        # Always isotropic resampling, so the first entry is ok.
        scale_name = 'scale' + (str(self.params.process.scale_text[scale][0])).replace('.', 'dot')
        if hasattr(self.params.radiomics, "scale_name"):
            setattr(self.params.radiomics, 'scale_name', scale_name)
        else:
            self.params.radiomics.scale_name = scale_name

        # Discretisation name
        gray_levels_name = (str(self.params.process.gray_levels[algo][gl])).replace('.', 'dot')

        if 'FBS' in self.params.process.algo[algo]:  # The minimum value defines the computation.
            if type(self.params.process.user_set_min_value) is list and self.params.process.user_set_min_value:
                min_val_name = '_min' + ((str(self.params.process.user_set_min_value)).replace('.', 'dot')).replace('-', 'M')
            else:
                # Otherwise, minimum value of ROI will be used (not recommended),
                # so no need to report it.
                min_val_name = ''
        else:
            min_val_name = ''

        if 'equal'in self.params.process.algo[algo]:
            # The number of gray-levels used for equalization is currently
            # hard-coded to 64 in equalization.m
            discretisation_name = 'algo' + self.params.process.algo[algo] + '256_bin' + gray_levels_name + min_val_name
        else:
            discretisation_name = 'algo' + self.params.process.algo[algo] + '_bin' + gray_levels_name + min_val_name

        # Processing full name
        processing_name = scale_name + '_' + discretisation_name
        if hasattr(self.params.radiomics, "processing_name"):
            setattr(self.params.radiomics, 'processing_name', processing_name)
        else:
            self.params.radiomics.processing_name = processing_name

    def init_from_nifti(self, nifti_image_path: Path) -> None:
        """Initializes the MEDscan class using a NIfTI file.

        Args:
            nifti_image_path (Path): NIfTI file path.

        Returns:
            None.
        
        """
        self.patientID = os.path.basename(nifti_image_path).split("_")[0]
        self.type = os.path.basename(nifti_image_path).split(".")[-3]
        self.format = "nifti"
        self.data.set_orientation(orientation="Axial")
        self.data.set_patient_position(patient_position="HFS")
        self.data.ROI.get_roi_from_path(roi_path=os.path.dirname(nifti_image_path), 
                                        id=Path(nifti_image_path).name.split("(")[0])
        self.data.volume.array = nib.load(nifti_image_path).get_fdata()
        # RAS to LPS
        self.data.volume.convert_to_LPS()
        self.data.volume.scan_rot = None
    
    def update_radiomics(
                        self, int_vol_hist_features: Dict = {}, 
                        morph_features: Dict = {}, loc_int_features: Dict = {}, 
                        stats_features: Dict = {}, int_hist_features: Dict = {},
                        glcm_features: Dict = {}, glrlm_features: Dict = {},
                        glszm_features: Dict = {}, gldzm_features: Dict = {}, 
                        ngtdm_features: Dict = {}, ngldm_features: Dict = {}) -> None:
        """Updates the results attribute with the extracted features.

        Args:
            int_vol_hist_features(Dict, optional): Dictionary of the intensity volume histogram features.
            morph_features(Dict, optional): Dictionary of the morphological features.
            loc_int_features(Dict, optional): Dictionary of the intensity local intensity features.
            stats_features(Dict, optional): Dictionary of the statistical features.
            int_hist_features(Dict, optional): Dictionary of the intensity histogram features.
            glcm_features(Dict, optional): Dictionary of the GLCM features.
            glrlm_features(Dict, optional): Dictionary of the GLRLM features.
            glszm_features(Dict, optional): Dictionary of the GLSZM features.
            gldzm_features(Dict, optional): Dictionary of the GLDZM features.
            ngtdm_features(Dict, optional): Dictionary of the NGTDM features.
            ngldm_features(Dict, optional): Dictionary of the NGLDM features.
        Returns:
            None.
        """
        # check glcm merge method
        glcm_merge_method = self.params.radiomics.glcm.merge_method
        if glcm_merge_method:
            if glcm_merge_method == 'average':
                glcm_merge_method = '_avg'
            elif glcm_merge_method == 'vol_merge':
                glcm_merge_method = '_comb'

        # check glrlm merge method
        glrlm_merge_method = self.params.radiomics.glrlm.merge_method
        if glrlm_merge_method:
            if glrlm_merge_method == 'average':
                glrlm_merge_method = '_avg'
            elif glrlm_merge_method == 'vol_merge':
                glrlm_merge_method = '_comb'

        # Non-texture Features
        if int_vol_hist_features:
            self.radiomics.image['intVolHist_3D'][self.params.radiomics.ivh_name] = int_vol_hist_features
        if morph_features:
            self.radiomics.image['morph_3D'][self.params.radiomics.scale_name] = morph_features
        if loc_int_features:
            self.radiomics.image['locInt_3D'][self.params.radiomics.scale_name] = loc_int_features
        if stats_features:
            self.radiomics.image['stats_3D'][self.params.radiomics.scale_name] = stats_features
        if int_hist_features:
            self.radiomics.image['intHist_3D'][self.params.radiomics.ih_name] = int_hist_features
        
        # Texture Features
        if glcm_features:
            self.radiomics.image['texture'][
                'glcm_3D' + glcm_merge_method][self.params.radiomics.processing_name] = glcm_features
        if glrlm_features:
            self.radiomics.image['texture'][
                'glrlm_3D' + glrlm_merge_method][self.params.radiomics.processing_name] = glrlm_features
        if glszm_features:
            self.radiomics.image['texture']['glszm_3D'][self.params.radiomics.processing_name] = glszm_features
        if gldzm_features:
            self.radiomics.image['texture']['gldzm_3D'][self.params.radiomics.processing_name] = gldzm_features
        if ngtdm_features:
            self.radiomics.image['texture']['ngtdm_3D'][self.params.radiomics.processing_name] = ngtdm_features
        if ngldm_features:
            self.radiomics.image['texture']['ngldm_3D'][self.params.radiomics.processing_name] = ngldm_features

    def save_radiomics(
                    self, scan_file_name: List, 
                    path_save: Path, roi_type: str, 
                    roi_type_label: str, patient_num: int = None) -> None:
        """
        Saves extracted radiomics features in a JSON file.

        Args:
            scan_file_name(List): List of scan files.
            path_save(Path): Saving path.
            roi_type(str): Type of the ROI.
            roi_type_label(str): Label of the ROI type.
            patient_num(int): Index of scan.
        
        Returns:
            None.
        """
        if path_save.name != f'features({roi_type})':
            if not (path_save / f'features({roi_type})').exists():
                (path_save / f'features({roi_type})').mkdir()
                path_save = Path(path_save / f'features({roi_type})')
            else:
                path_save = Path(path_save) / f'features({roi_type})'
        else:
            path_save = Path(path_save)
        params = {}
        params['roi_type'] = roi_type_label
        params['patientID'] = self.patientID
        params['vox_dim'] = list([
                                self.data.volume.spatialRef.PixelExtentInWorldX, 
                                self.data.volume.spatialRef.PixelExtentInWorldY,
                                self.data.volume.spatialRef.PixelExtentInWorldZ
                                ])
        self.radiomics.update_params(params)
        if type(scan_file_name) is str:
            index_dot = scan_file_name.find('.')
            ext = scan_file_name.find('.npy')
            name_save = scan_file_name[:index_dot] + \
                        '(' + roi_type_label + ')' + \
                        scan_file_name[index_dot : ext]
        elif patient_num is not None:
            index_dot = scan_file_name[patient_num].find('.')
            ext = scan_file_name[patient_num].find('.npy')
            name_save = scan_file_name[patient_num][:index_dot] + \
                        '(' + roi_type_label + ')' + \
                        scan_file_name[patient_num][index_dot : ext]
        else:
            raise ValueError("`patient_num` must be specified or `scan_file_name` must be str")

        with open(path_save / f"{name_save}.json", "w") as fp:   
            dump(self.radiomics.to_json(), fp, indent=4, cls=NumpyEncoder)


    class Params:
        """Organizes all processing, filtering and features extraction parameters"""

        def __init__(self) -> None:
            """
            Organizes all processing, filtering and features extraction
            """
            self.process = self.Process()
            self.filter = self.Filter()
            self.radiomics = self.Radiomics()


        class Process:
            """Organizes all processing parameters."""
            def __init__(self, **kwargs) -> None:
                """
                Constructor of the `Process` class.
                """
                self.algo = kwargs['algo'] if 'algo' in kwargs else None
                self.box_string = kwargs['box_string'] if 'box_string' in kwargs else None
                self.gl_round = kwargs['gl_round'] if 'gl_round' in kwargs else None
                self.gray_levels = kwargs['gray_levels'] if 'gray_levels' in kwargs else None
                self.ih = kwargs['ih'] if 'ih' in kwargs else None
                self.im_range = kwargs['im_range'] if 'im_range' in kwargs else None
                self.im_type = kwargs['im_type'] if 'im_type' in kwargs else None
                self.intensity_type = kwargs['intensity_type'] if 'intensity_type' in kwargs else None
                self.ivh = kwargs['ivh'] if 'ivh' in kwargs else None
                self.n_algo = kwargs['n_algo'] if 'n_algo' in kwargs else None
                self.n_exp = kwargs['n_exp'] if 'n_exp' in kwargs else None
                self.n_gl = kwargs['n_gl'] if 'n_gl' in kwargs else None
                self.n_scale = kwargs['n_scale'] if 'n_scale' in kwargs else None
                self.outliers = kwargs['outliers'] if 'outliers' in kwargs else None
                self.scale_non_text = kwargs['scale_non_text'] if 'scale_non_text' in kwargs else None
                self.scale_text = kwargs['scale_text'] if 'scale_text' in kwargs else None
                self.roi_interp = kwargs['roi_interp'] if 'roi_interp' in kwargs else None
                self.roi_pv = kwargs['roi_pv'] if 'roi_pv' in kwargs else None
                self.user_set_min_value = kwargs['user_set_min_value'] if 'user_set_min_value' in kwargs else None
                self.vol_interp = kwargs['vol_interp'] if 'vol_interp' in kwargs else None

            def init_from_json(self, path_to_json: Union[Path, str]) -> None:
                """
                Updates class attributes from json file.

                Args:
                    path_to_json(Union[Path, str]): Path to the JSON file with processing parameters.
                
                Returns:
                    None.
                """
                __params = load_json(Path(path_to_json))

                self.algo = __params['algo'] if 'algo' in __params else self.algo
                self.box_string = __params['box_string'] if 'box_string' in __params else self.box_string
                self.gl_round = __params['gl_round'] if 'gl_round' in __params else self.gl_round
                self.gray_levels = __params['gray_levels'] if 'gray_levels' in __params else self.gray_levels
                self.ih = __params['ih'] if 'ih' in __params else self.ih
                self.im_range = __params['im_range'] if 'im_range' in __params else self.im_range
                self.im_type = __params['im_type'] if 'im_type' in __params else self.im_type
                self.ivh = __params['ivh'] if 'ivh' in __params else self.ivh
                self.n_algo = __params['n_algo'] if 'n_algo' in __params else self.n_algo
                self.n_exp = __params['n_exp'] if 'n_exp' in __params else self.n_exp
                self.n_gl = __params['n_gl'] if 'n_gl' in __params else self.n_gl
                self.n_scale = __params['n_scale'] if 'n_scale' in __params else self.n_scale
                self.outliers = __params['outliers'] if 'outliers' in __params else self.outliers
                self.scale_non_text = __params['scale_non_text'] if 'scale_non_text' in __params else self.scale_non_text
                self.scale_text = __params['scale_text'] if 'scale_text' in __params else self.scale_text
                self.roi_interp = __params['roi_interp'] if 'roi_interp' in __params else self.roi_interp
                self.roi_pv = __params['roi_pv'] if 'roi_pv' in __params else self.roi_pv
                self.user_set_min_value = __params['user_set_min_value'] if 'user_set_min_value' in __params else self.user_set_min_value
                self.vol_interp = __params['vol_interp'] if 'vol_interp' in __params else self.vol_interp


        class Filter:
            """Organizes all filtering parameters"""
            def __init__(self, filter_type: str = "") -> None:
                """
                Constructor of the Filter class.

                Args:
                    filter_type(str): Type of the filter that will be used (Must be 'mean', 'log', 'laws',
                        'gabor' or 'wavelet').
                
                Returns:
                    None.
                """
                self.filter_type = filter_type
                self.mean = self.Mean()
                self.log = self.Log()
                self.gabor = self.Gabor()
                self.laws = self.Laws()
                self.wavelet = self.Wavelet()
                self.textural = self.Textural()


            class Mean:
                """Organizes the Mean filter parameters"""
                def __init__(
                        self, ndims: int = 0, name_save: str = '', 
                        padding: str = '', size: int = 0, orthogonal_rot: bool = False
                ) -> None:
                    """
                    Constructor of the Mean class.

                    Args:
                        ndims(int): Filter dimension. 
                        name_save(str): Specific name added to final extraction results file. 
                        padding(str): padding mode. 
                        size(int): Filter size. 
                                       
                    Returns:
                        None.
                    """
                    self.name_save = name_save
                    self.ndims = ndims
                    self.orthogonal_rot = orthogonal_rot
                    self.padding = padding
                    self.size = size

                def init_from_json(self, params: Dict) -> None:
                    """
                    Updates class attributes from json file.

                    Args:
                        params(Dict): Dictionary of the Mean filter parameters.
                    
                    Returns:
                        None.
                    """
                    self.name_save = params['name_save']
                    self.ndims = params['ndims']
                    self.padding = params['padding']
                    self.size = params['size']
                    self.orthogonal_rot = params['orthogonal_rot']


            class Log:
                """Organizes the Log filter parameters"""
                def __init__(
                        self, ndims: int = 0, sigma: float = 0.0, 
                        padding: str = '', orthogonal_rot: bool = False, 
                        name_save: str = ''
                ) -> None:
                    """
                    Constructor of the Log class.

                    Args:
                        ndims(int): Filter dimension. 
                        sigma(float): Float of the sigma value.
                        padding(str): padding mode.
                        orthogonal_rot(bool): If True will compute average response over orthogonal planes. 
                        name_save(str): Specific name added to final extraction results file. 
                                       
                    Returns:
                        None.
                    """
                    self.name_save = name_save
                    self.ndims = ndims
                    self.orthogonal_rot = orthogonal_rot
                    self.padding = padding
                    self.sigma = sigma

                def init_from_json(self, params: Dict) -> None:
                    """
                    Updates class attributes from json file.

                    Args:
                        params(Dict): Dictionary of the Log filter parameters.
                    
                    Returns:
                        None.
                    """
                    self.name_save = params['name_save']
                    self.ndims = params['ndims']
                    self.orthogonal_rot = params['orthogonal_rot']
                    self.padding = params['padding']
                    self.sigma = params['sigma']


            class Gabor:
                """Organizes the gabor filter parameters"""
                def __init__(
                        self, sigma: float = 0.0, _lambda: float = 0.0,  
                        gamma: float = 0.0, theta: str = '', rot_invariance: bool = False,
                        orthogonal_rot: bool= False, name_save: str = '',
                        padding: str = ''
                ) -> None:
                    """
                    Constructor of the Gabor class.

                    Args:
                        sigma(float): Float of the sigma value.
                        _lambda(float): Float of the lambda value.
                        gamma(float): Float of the gamma value.
                        theta(str): String of the theta angle value.
                        rot_invariance(bool): If True the filter will be rotation invariant.
                        orthogonal_rot(bool): If True will compute average response over orthogonal planes.
                        name_save(str): Specific name added to final extraction results file.
                        padding(str): padding mode.
                                       
                    Returns:
                        None.
                    """
                    self._lambda = _lambda
                    self.gamma = gamma
                    self.name_save = name_save
                    self.orthogonal_rot = orthogonal_rot
                    self.padding = padding
                    self.rot_invariance = rot_invariance
                    self.sigma = sigma
                    self.theta = theta

                def init_from_json(self, params: Dict) -> None:
                    """
                    Updates class attributes from json file.

                    Args:
                        params(Dict): Dictionary of the gabor filter parameters.
                    
                    Returns:
                        None.
                    """
                    self._lambda = params['lambda']
                    self.gamma = params['gamma']
                    self.name_save = params['name_save']
                    self.orthogonal_rot = params['orthogonal_rot']
                    self.padding = params['padding']
                    self.rot_invariance = params['rot_invariance']
                    self.sigma = params['sigma']
                    if type(params["theta"]) is str:
                        if params["theta"].lower().startswith('pi/'):
                            self.theta = np.pi / int(params["theta"].split('/')[1])
                        elif params["theta"].lower().startswith('-'):
                            if params["theta"].lower().startswith('-pi/'):
                                self.theta = -np.pi / int(params["theta"].split('/')[1])
                            else:
                                nom, denom = params["theta"].replace('-', '').replace('Pi', '').split('/')
                                self.theta = -np.pi*int(nom) / int(denom)
                    else:
                        self.theta = float(params["theta"])


            class Laws:
                """Organizes the laws filter parameters"""
                def __init__(
                        self, config: List = [], energy_distance: int = 0, 
                        energy_image: bool = False, rot_invariance: bool = False, 
                        orthogonal_rot: bool = False, name_save: str = '', padding: str = ''
                ) -> None:
                    """
                    Constructor of the Laws class.

                    Args:
                        config(List): Configuration of the Laws filter, for ex: ['E5', 'L5', 'E5'].
                        energy_distance(int): Chebyshev distance.
                        energy_image(bool): If True will compute the Laws texture energy image.
                        rot_invariance(bool): If True the filter will be rotation invariant.
                        orthogonal_rot(bool): If True will compute average response over orthogonal planes.
                        name_save(str): Specific name added to final extraction results file.
                        padding(str): padding mode.
                                       
                    Returns:
                        None.
                    """
                    self.config = config
                    self.energy_distance = energy_distance
                    self.energy_image = energy_image
                    self.name_save = name_save
                    self.orthogonal_rot = orthogonal_rot
                    self.padding = padding
                    self.rot_invariance = rot_invariance

                def init_from_json(self, params: Dict) -> None:
                    """
                    Updates class attributes from json file.

                    Args:
                        params(Dict): Dictionary of the laws filter parameters.
                    
                    Returns:
                        None.
                    """
                    self.config = params['config']
                    self.energy_distance = params['energy_distance']
                    self.energy_image = params['energy_image']
                    self.name_save = params['name_save']
                    self.orthogonal_rot = params['orthogonal_rot']
                    self.padding = params['padding']
                    self.rot_invariance = params['rot_invariance']


            class Wavelet:
                """Organizes the Wavelet filter parameters"""
                def __init__(
                        self, ndims: int = 0, name_save: str = '', 
                        basis_function: str = '', subband: str = '', level: int = 0, 
                        rot_invariance: bool = False, padding: str = ''
                ) -> None:
                    """
                    Constructor of the Wavelet class.

                    Args:
                        ndims(int): Dimension of the filter.
                        name_save(str): Specific name added to final extraction results file.
                        basis_function(str): Wavelet basis function.
                        subband(str): Wavelet subband.
                        level(int): Decomposition level.
                        rot_invariance(bool): If True the filter will be rotation invariant.
                        padding(str): padding mode.
                                       
                    Returns:
                        None.
                    """
                    self.basis_function = basis_function
                    self.level = level
                    self.ndims = ndims
                    self.name_save = name_save
                    self.padding = padding
                    self.rot_invariance = rot_invariance
                    self.subband = subband

                def init_from_json(self, params: Dict) -> None:
                    """
                    Updates class attributes from json file.

                    Args:
                        params(Dict): Dictionary of the wavelet filter parameters.
                    
                    Returns:
                        None.
                    """
                    self.basis_function = params['basis_function']
                    self.level = params['level']
                    self.ndims = params['ndims']
                    self.name_save = params['name_save']
                    self.padding = params['padding']
                    self.rot_invariance = params['rot_invariance']
                    self.subband = params['subband']
            

            class Textural:
                """Organizes the Textural filters parameters"""
                def __init__(
                        self,
                        family: str = '',
                        size: int = 0,
                        discretization: dict = {},
                        local: bool = False,
                        name_save: str = ''
                ) -> None:
                    """
                    Constructor of the Textural class.

                    Args:
                        family (str, optional): The family of the textural filter.
                        size (int, optional): The filter size.
                        discretization (dict, optional): The discretization parameters.
                        local (bool, optional): If true, the discretization will be computed locally, else globally.
                        name_save (str, optional): Specific name added to final extraction results file.
                                       
                    Returns:
                        None.
                    """
                    self.family = family
                    self.size = size
                    self.discretization = discretization
                    self.local = local
                    self.name_save = name_save

                def init_from_json(self, params: Dict) -> None:
                    """
                    Updates class attributes from json file.

                    Args:
                        params(Dict): Dictionary of the wavelet filter parameters.
                    
                    Returns:
                        None.
                    """
                    self.family = params['family']
                    self.size = params['size']
                    self.discretization = params['discretization']
                    self.local = params['local']
                    self.name_save = params['name_save']


        class Radiomics:
            """Organizes the radiomics extraction parameters"""
            def __init__(self, **kwargs) -> None:
                """
                Constructor of the Radiomics class.
                """
                self.ih_name = kwargs['ih_name'] if 'ih_name' in kwargs else None
                self.ivh_name = kwargs['ivh_name'] if 'ivh_name' in kwargs else None
                self.glcm = self.GLCM()
                self.glrlm = self.GLRLM()
                self.ngtdm = self.NGTDM()
                self.name_text_types = kwargs['name_text_types'] if 'name_text_types' in kwargs else None
                self.processing_name = kwargs['processing_name'] if 'processing_name' in kwargs else None
                self.scale_name = kwargs['scale_name'] if 'scale_name' in kwargs else None
                self.extract = kwargs['extract'] if 'extract' in kwargs else {}

            class GLCM:
                """Organizes the GLCM features extraction parameters"""
                def __init__(
                        self, 
                        dist_correction: Union[bool, str] = False,
                        merge_method: str = "vol_merge"
                ) -> None:
                    """
                    Constructor of the GLCM class

                    Args:
                        dist_correction(Union[bool, str]): norm for distance weighting, must be 
                            "manhattan", "euclidean" or "chebyshev". If True the norm for distance weighting 
                            is gonna be "euclidean".
                        merge_method(str): merging method which determines how features are
                            calculated. Must be "average", "slice_merge", "dir_merge" and "vol_merge".
                                       
                    Returns:
                        None.
                    """
                    self.dist_correction = dist_correction
                    self.merge_method = merge_method


            class GLRLM:
                """Organizes the GLRLM features extraction parameters"""
                def __init__(
                        self, 
                        dist_correction: Union[bool, str] = False,
                        merge_method: str = "vol_merge"
                ) -> None:
                    """
                    Constructor of the GLRLM class

                    Args:
                        dist_correction(Union[bool, str]): If True the norm for distance weighting is gonna be "euclidean".
                        merge_method(str): merging method which determines how features are
                            calculated. Must be "average", "slice_merge", "dir_merge" and "vol_merge".

                    Returns:
                        None.
                    """
                    self.dist_correction = dist_correction
                    self.merge_method = merge_method


            class NGTDM:
                """Organizes the NGTDM features extraction parameters"""
                def __init__(
                        self, 
                        dist_correction: Union[bool, str] = None
                ) -> None:
                    """
                    Constructor of the NGTDM class

                    Args:
                        dist_correction(Union[bool, str]): If True the norm for distance weighting is gonna be "euclidean".

                    Returns:
                        None.
                    """
                    self.dist_correction = dist_correction


    class Radiomics:
        """Organizes all the extracted features.
        """
        def __init__(self, image: Dict = None, params: Dict = None) -> None:
            """Constructor of the Radiomics class
            Args:
                image(Dict): Dict of the extracted features.
                params(Dict): Dict of the parameters used in features extraction (roi type, voxels diemension...)
            
            Returns:
                None
            """
            self.image = image if image else {}
            self.params = params if params else {}

        def update_params(self, params: Dict) -> None:
            """Updates `params` attribute from a given Dict
                Args:
                    params(Dict): Dict of the parameters used in features extraction (roi type, voxels diemension...)
                
                Returns:
                    None
            """
            self.params['roi_type'] = params['roi_type']
            self.params['patientID'] = params['patientID']
            self.params['vox_dim'] = params['vox_dim']

        def to_json(self) -> Dict:
            """Summarizes the class attributes in a Dict
                Args:
                    None
                
                Returns:
                    Dict: Dictionay of radiomics structure (extracted features and extraction params)
            """
            radiomics = {
                'image': self.image,
                'params': self.params
            }
            return radiomics


    class data:
        """Organizes all imaging data (volume and ROI). 

        Attributes:
            volume (object): Instance of MEDscan.data.volume inner class.
            ROI (object): Instance of MEDscan.data.ROI inner class.
            orientation (str): Imaging data orientation (axial, sagittal or coronal).
            patient_position (str): Patient position specifies the position of the 
                patient relative to the imaging equipment space (HFS, HFP...).

        """
        def __init__(self, orientation: str=None, patient_position: str=None) -> None:
            """Constructor of the scan class

            Args:
                orientation (str, optional): Imaging data orientation (axial, sagittal or coronal).
                patient_position (str, optional): Patient position specifies the position of the 
                    patient relative to the imaging equipment space (HFS, HFP...).
            
            Returns:
                None.
            """
            self.volume = self.volume() 
            self.volume_process = self.volume_process()
            self.ROI = self.ROI()
            self.orientation = orientation
            self.patient_position = patient_position

        def set_patient_position(self, patient_position):
            self.patient_position = patient_position

        def set_orientation(self, orientation):
            self.orientation = orientation
        
        def set_volume(self, volume):
            self.volume = volume
        
        def set_ROI(self, *args):
            self.ROI = self.ROI(args)

        def get_roi_from_indexes(self, key: int) -> np.ndarray:
            """
            Extracts ROI data using the saved indexes (Indexes of non-null values).

            Args:
                key (int): Key of ROI indexes list (A volume can have multiple ROIs).

            Returns:
                ndarray: n-dimensional array of ROI data.
            
            """
            roi_volume = np.zeros_like(self.volume.array).flatten()
            roi_volume[self.ROI.get_indexes(key)] = 1
            return roi_volume.reshape(self.volume.array.shape)

        def get_indexes_by_roi_name(self, roi_name : str) -> np.ndarray:
            """
            Extract ROI data using the ROI name.

            Args:
                roi_name (str): String of the ROI name (A volume can have multiple ROIs).

            Returns:
                ndarray: n-dimensional array of the ROI data.
            
            """
            roi_name_key = list(self.ROI.roi_names.values()).index(roi_name)
            roi_volume = np.zeros_like(self.volume.array).flatten()
            roi_volume[self.ROI.get_indexes(roi_name_key)] = 1
            return roi_volume.reshape(self.volume.array.shape)

        def display(self, _slice: int = None, roi: Union[str, int] = 0) -> None:
            """Displays slices from imaging data with the ROI contour in XY-Plane.

            Args:
                _slice (int, optional): Index of the slice you want to plot.
                roi (Union[str, int], optional): ROI name or index. If not specified will use the first ROI.

            Returns:
                None.
            
            """
            # extract slices containing ROI
            size_m = self.volume.array.shape
            i = np.arange(0, size_m[0])
            j = np.arange(0, size_m[1])
            k = np.arange(0, size_m[2])
            ind_mask = np.nonzero(self.get_roi_from_indexes(roi))
            J, I, K = np.meshgrid(i, j, k, indexing='ij')
            I = I[ind_mask]
            J = J[ind_mask]
            K = K[ind_mask]
            slices = np.unique(K)

            vol_data = self.volume.array.swapaxes(0, 1)[:, :, slices]
            roi_data = self.get_roi_from_indexes(roi).swapaxes(0, 1)[:, :, slices]        
            
            rows = int(np.round(np.sqrt(len(slices))))
            columns = int(np.ceil(len(slices) / rows))
            
            plt.set_cmap(plt.gray())
            
            # plot only one slice
            if _slice:
                fig, ax =  plt.subplots(1, 1, figsize=(10, 5))
                ax.axis('off')
                ax.set_title(_slice)
                ax.imshow(vol_data[:, :, _slice])
                im = Image.fromarray((roi_data[:, :, _slice]))
                ax.contour(im, colors='red', linewidths=0.4, alpha=0.45)
                lps_ax = fig.add_subplot(1, columns, 1)
            
            # plot multiple slices containing an ROI.
            else:
                fig, axs =  plt.subplots(rows, columns+1, figsize=(20, 10))
                s = 0
                for i in range(0,rows):
                    for j in range(0,columns):
                        axs[i,j].axis('off')
                        if s < len(slices):
                            axs[i,j].set_title(str(s))
                            axs[i,j].imshow(vol_data[:, :, s])
                            im = Image.fromarray((roi_data[:, :, s]))
                            axs[i,j].contour(im, colors='red', linewidths=0.4, alpha=0.45)
                        s += 1
                    axs[i,columns].axis('off')
                lps_ax = fig.add_subplot(1, columns+1, axs.shape[1])

            fig.suptitle('XY-Plane')
            fig.tight_layout()
            
            # add the coordinates system
            lps_ax.axis([-1.5, 1.5, -1.5, 1.5])
            lps_ax.set_title("Coordinates system")
            
            lps_ax.quiver([-0.5], [0], [1.5], [0], scale_units='xy', angles='xy', scale=1.0, color='green')
            lps_ax.quiver([-0.5], [0], [0], [-1.5], scale_units='xy', angles='xy', scale=3, color='blue')
            lps_ax.quiver([-0.5], [0], [1.5], [1.5], scale_units='xy', angles='xy', scale=3, color='red')
            lps_ax.text(1.0, 0, "L")
            lps_ax.text(-0.3, -0.5, "P")
            lps_ax.text(0.3, 0.4, "S")

            lps_ax.set_xticks([])
            lps_ax.set_yticks([])

            plt.show()

        def display_process(self, _slice: int = None, roi: Union[str, int] = 0) -> None:
            """Displays slices from imaging data with the ROI contour in XY-Plane.

            Args:
                _slice (int, optional): Index of the slice you want to plot.
                roi (Union[str, int], optional): ROI name or index. If not specified will use the first ROI.

            Returns:
                None.
            
            """
            # extract slices containing ROI
            size_m = self.volume_process.array.shape
            i = np.arange(0, size_m[0])
            j = np.arange(0, size_m[1])
            k = np.arange(0, size_m[2])
            ind_mask = np.nonzero(self.get_roi_from_indexes(roi))
            J, I, K = np.meshgrid(j, i, k, indexing='ij')
            I = I[ind_mask]
            J = J[ind_mask]
            K = K[ind_mask]
            slices = np.unique(K)

            vol_data = self.volume_process.array.swapaxes(0, 1)[:, :, slices]
            roi_data = self.get_roi_from_indexes(roi).swapaxes(0, 1)[:, :, slices]        
            
            rows = int(np.round(np.sqrt(len(slices))))
            columns = int(np.ceil(len(slices) / rows))
            
            plt.set_cmap(plt.gray())
            
            # plot only one slice
            if _slice:
                fig, ax =  plt.subplots(1, 1, figsize=(10, 5))
                ax.axis('off')
                ax.set_title(_slice)
                ax.imshow(vol_data[:, :, _slice])
                im = Image.fromarray((roi_data[:, :, _slice]))
                ax.contour(im, colors='red', linewidths=0.4, alpha=0.45)
                lps_ax = fig.add_subplot(1, columns, 1)
            
            # plot multiple slices containing an ROI.
            else:
                fig, axs =  plt.subplots(rows, columns+1, figsize=(20, 10))
                s = 0
                for i in range(0,rows):
                    for j in range(0,columns):
                        axs[i,j].axis('off')
                        if s < len(slices):
                            axs[i,j].set_title(str(s))
                            axs[i,j].imshow(vol_data[:, :, s])
                            im = Image.fromarray((roi_data[:, :, s]))
                            axs[i,j].contour(im, colors='red', linewidths=0.4, alpha=0.45)
                        s += 1
                    axs[i,columns].axis('off')
                lps_ax = fig.add_subplot(1, columns+1, axs.shape[1])

            fig.suptitle('XY-Plane')
            fig.tight_layout()
            
            # add the coordinates system
            lps_ax.axis([-1.5, 1.5, -1.5, 1.5])
            lps_ax.set_title("Coordinates system")
            
            lps_ax.quiver([-0.5], [0], [1.5], [0], scale_units='xy', angles='xy', scale=1.0, color='green')
            lps_ax.quiver([-0.5], [0], [0], [-1.5], scale_units='xy', angles='xy', scale=3, color='blue')
            lps_ax.quiver([-0.5], [0], [1.5], [1.5], scale_units='xy', angles='xy', scale=3, color='red')
            lps_ax.text(1.0, 0, "L")
            lps_ax.text(-0.3, -0.5, "P")
            lps_ax.text(0.3, 0.4, "S")

            lps_ax.set_xticks([])
            lps_ax.set_yticks([])

            plt.show()


        class volume:
            """Organizes all volume data and information related to imaging volume. 

            Attributes:
                spatialRef (imref3d): Imaging data orientation (axial, sagittal or coronal).
                scan_rot (ndarray): Array of the rotation applied to the XYZ points of the ROI.
                array (ndarray): n-dimensional of the imaging data.

            """
            def __init__(self, spatialRef: imref3d=None, scan_rot: str=None, array: np.ndarray=None) -> None:
                """Organizes all volume data and information. 

                Args:
                    spatialRef (imref3d, optional): Imaging data orientation (axial, sagittal or coronal).
                    scan_rot (ndarray, optional): Array of the rotation applied to the XYZ points of the ROI.
                    array (ndarray, optional): n-dimensional of the imaging data.

                """
                self.spatialRef = spatialRef
                self.scan_rot = scan_rot
                self.array = array

            def update_spatialRef(self, spatialRef_value):
                self.spatialRef = spatialRef_value
            
            def update_scan_rot(self, scan_rot_value):
                self.scan_rot = scan_rot_value
            
            def update_transScanToModel(self, transScanToModel_value):
                self.transScanToModel = transScanToModel_value
            
            def update_array(self, array):
                self.array = array

            def convert_to_LPS(self):
                """Convert Imaging data to LPS (Left-Posterior-Superior) coordinates system.
                <https://www.slicer.org/wiki/Coordinate_systems>.

                Returns:
                    None.

                """
                # flip x
                self.array = np.flip(self.array, 0)
                # flip y
                self.array = np.flip(self.array, 1)
            
            def spatialRef_from_nifti(self, nifti_image_path: Union[Path, str]) -> None:
                """Computes the imref3d spatialRef using a NIFTI file and
                updates the `spatialRef` attribute.

                Args:
                    nifti_image_path (str): String of the NIFTI file path.

                Returns:
                    None.

                """
                # Loading the nifti file:
                nifti_image_path = Path(nifti_image_path)
                nifti = nib.load(nifti_image_path)
                nifti_data = self.array

                # spatialRef Creation
                pixelX = nifti.affine[0, 0]
                pixelY = nifti.affine[1, 1]
                sliceS = nifti.affine[2, 2]
                min_grid = nifti.affine[:3, 3]
                min_Xgrid = min_grid[0]
                min_Ygrid = min_grid[1]
                min_Zgrid = min_grid[2]
                size_image = np.shape(nifti_data)
                spatialRef = imref3d(size_image, abs(pixelX), abs(pixelY), abs(sliceS))
                spatialRef.XWorldLimits = (np.array(spatialRef.XWorldLimits) -
                                        (spatialRef.XWorldLimits[0] -
                                            (min_Xgrid-pixelX/2))
                                        ).tolist()
                spatialRef.YWorldLimits = (np.array(spatialRef.YWorldLimits) -
                                        (spatialRef.YWorldLimits[0] -
                                            (min_Ygrid-pixelY/2))
                                        ).tolist()
                spatialRef.ZWorldLimits = (np.array(spatialRef.ZWorldLimits) -
                                        (spatialRef.ZWorldLimits[0] -
                                            (min_Zgrid-sliceS/2))
                                        ).tolist()

                # Converting the results into lists
                spatialRef.ImageSize = spatialRef.ImageSize.tolist()
                spatialRef.XIntrinsicLimits = spatialRef.XIntrinsicLimits.tolist()
                spatialRef.YIntrinsicLimits = spatialRef.YIntrinsicLimits.tolist()
                spatialRef.ZIntrinsicLimits = spatialRef.ZIntrinsicLimits.tolist()

                # update spatialRef
                self.update_spatialRef(spatialRef)

            def convert_spatialRef(self):
                """converts the `spatialRef` attribute from RAS to LPS coordinates system.
                <https://www.slicer.org/wiki/Coordinate_systems>.

                Args:
                    None.

                Returns:
                    None.

                """
                # swap x and y data
                temp = self.spatialRef.ImageExtentInWorldX
                self.spatialRef.ImageExtentInWorldX = self.spatialRef.ImageExtentInWorldY
                self.spatialRef.ImageExtentInWorldY = temp

                temp = self.spatialRef.PixelExtentInWorldX
                self.spatialRef.PixelExtentInWorldX = self.spatialRef.PixelExtentInWorldY
                self.spatialRef.PixelExtentInWorldY = temp

                temp = self.spatialRef.XIntrinsicLimits
                self.spatialRef.XIntrinsicLimits = self.spatialRef.YIntrinsicLimits
                self.spatialRef.YIntrinsicLimits = temp

                temp = self.spatialRef.XWorldLimits
                self.spatialRef.XWorldLimits = self.spatialRef.YWorldLimits
                self.spatialRef.YWorldLimits = temp
                del temp

        class volume_process:
            """Organizes all volume data and information. 

            Attributes:
                spatialRef (imref3d): Imaging data orientation (axial, sagittal or coronal).
                scan_rot (ndarray): Array of the rotation applied to the XYZ points of the ROI.
                data (ndarray): n-dimensional of the imaging data.

            """
            def __init__(self, spatialRef: imref3d = None, 
                        scan_rot: List = None, array: np.ndarray = None,
                        user_string: str = "") -> None:
                """Organizes all volume data and information. 

                Args:
                    spatialRef (imref3d, optional): Imaging data orientation (axial, sagittal or coronal).
                    scan_rot (ndarray, optional): Array of the rotation applied to the XYZ points of the ROI.
                    array (ndarray, optional): n-dimensional of the imaging data.
                    user_string(str, optional): string explaining the processed data in the class.
                
                Returns:
                    None.

                """
                self.array = array
                self.scan_rot = scan_rot
                self.spatialRef = spatialRef
                self.user_string = user_string

            def update_processed_data(self, array: np.ndarray, user_string: str = "") -> None:
                if user_string:
                    self.user_string = user_string
                self.array = array

            def save(self, name_save: str, path_save: Union[Path, str])-> None:
                """Saves the processed data locally.

                Args:
                    name_save(str): Saving name of the processed data.
                    path_save(Union[Path, str]): Path to where save the processed data.
                
                Returns:
                    None.
                """
                path_save = Path(path_save)
                if not name_save:
                    name_save = self.user_string

                if not name_save.endswith('.npy'):
                    name_save += '.npy'

                with open(path_save / name_save, 'wb') as f:
                    np.save(f, self.array)

            def load(
                    self, 
                    file_name: str, 
                    loading_path: Union[Path, str], 
                    update: bool=True
                ) -> Union[None, np.ndarray]:
                """Saves the processed data locally.

                Args:
                    file_name(str): Name file of the processed data to load.
                    loading_path(Union[Path, str]): Path to the processed data to load.
                    update(bool, optional): If True, updates the class attrtibutes with loaded data.

                Returns:
                    None.
                """
                loading_path = Path(loading_path)

                if not file_name.endswith('.npy'):
                    file_name += '.npy'

                with open(loading_path / file_name, 'rb') as f:
                    if update:
                        self.update_processed_data(np.load(f, allow_pickle=True))
                    else:
                        return np.load(f, allow_pickle=True)


        class ROI:
            """Organizes all ROI data and information. 

            Attributes:
                indexes (Dict): Dict of the ROI indexes for each ROI name.
                roi_names (Dict): Dict of the ROI names.
                nameSet (Dict): Dict of the User-defined name for Structure Set for each ROI name.
                nameSetInfo (Dict): Dict of the names of the structure sets that define the areas of 
                    significance. Either 'StructureSetName', 'StructureSetDescription', 'SeriesDescription' 
                    or 'SeriesInstanceUID'.
            
            """
            def __init__(self, indexes: Dict=None, roi_names: Dict=None) -> None:
                """Constructor of the ROI class.

                Args:
                    indexes (Dict, optional): Dict of the ROI indexes for each ROI name.
                    roi_names (Dict, optional): Dict of the ROI names.
                
                Returns:
                    None.
                """
                self.indexes = indexes if indexes else {}
                self.roi_names = roi_names if roi_names else {}
                self.nameSet = roi_names if roi_names else {}
                self.nameSetInfo = roi_names if roi_names else {}

            def get_indexes(self, key):
                if not self.indexes or key is None:
                    return {}
                else:
                    return self.indexes[str(key)]

            def get_roi_name(self, key):
                if not self.roi_names or key is None:
                    return {}
                else:
                    return self.roi_names[str(key)]

            def get_name_set(self, key):
                if not self.nameSet or key is None:
                    return {}
                else:
                    return self.nameSet[str(key)]

            def get_name_set_info(self, key):
                if not self.nameSetInfo or key is None:
                    return {}
                else:
                    return self.nameSetInfo[str(key)]

            def update_indexes(self, key, indexes):
                try: 
                    self.indexes[str(key)] = indexes
                except:
                    Warning.warn("Wrong key given in update_indexes()")

            def update_roi_name(self, key, roi_name):
                try:
                    self.roi_names[str(key)] = roi_name
                except:
                    Warning.warn("Wrong key given in update_roi_name()")

            def update_name_set(self, key, name_set):
                try:
                    self.nameSet[str(key)] = name_set
                except:
                    Warning.warn("Wrong key given in update_name_set()")

            def update_name_set_info(self, key, nameSetInfo):
                try:
                    self.nameSetInfo[str(key)] = nameSetInfo
                except:
                    Warning.warn("Wrong key given in update_name_set_info()")
            
            def convert_to_LPS(self, data: np.ndarray) -> np.ndarray:
                """Converts the given volume to LPS coordinates system. For 
                more details please refer here : https://www.slicer.org/wiki/Coordinate_systems
                Args:
                    data(ndarray) : Volume data in RAS to convert to to LPS

                Returns:
                    ndarray: n-dimensional of `data` in LPS.
                """
                # flip x
                data = np.flip(data, 0)
                # flip y
                data = np.flip(data, 1)

                return data

            def get_roi_from_path(self, roi_path: Union[Path, str], id: str):
                """Extracts all ROI data from the given path for the given
                patient ID and updates all class attributes with the new extracted data.

                Args:
                    roi_path(Union[Path, str]): Path where the ROI data is stored.
                    id(str): ID containing patient ID and the modality type, to identify the right file.
                
                Returns:
                    None.
                """
                self.indexes = {}
                self.roi_names = {}
                self.nameSet = {}
                self.nameSetInfo = {}
                roi_index = 0
                list_of_patients = os.listdir(roi_path)

                for file in list_of_patients:
                    # Load the patient's ROI nifti files :
                    if file.startswith(id) and file.endswith('nii.gz') and 'ROI' in file.split("."):
                        roi = nib.load(roi_path + "/" + file)
                        roi_data = self.convert_to_LPS(data=roi.get_fdata())
                        roi_name = file[file.find("(")+1 : file.find(")")]
                        name_set = file[file.find("_")+2 : file.find("(")]
                        self.update_indexes(key=roi_index, indexes=np.nonzero(roi_data.flatten()))
                        self.update_name_set(key=roi_index, name_set=name_set)
                        self.update_roi_name(key=roi_index, roi_name=roi_name)
                        roi_index += 1



import logging

from . import utils
from . import processing
from . import biomarkers
from . import filters
from . import wrangling
from . import learning
from .MEDscan import MEDscan


stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.WARNING)
logging.getLogger(__name__).addHandler(stream_handler)

__author__ = "MEDomicsLab consortium"
__version__ = "0.9.10"
__copyright__ = "Copyright (C) MEDomicsLab consortium"
__license__ = "GNU General Public License 3.0"
__maintainer__ = "MAHDI AIT LHAJ LOUTFI"
__email__ = "medomics.info@gmail.com"
